#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
from collections.abc import Iterator

from robot.utils import normalize_whitespace

from .context import (FileContext, KeywordContext, LexingContext, SuiteFileContext,
                      TestCaseContext)
from .statementlexers import (BreakLexer, CommentLexer, CommentSectionHeaderLexer,
                              ContinueLexer, ElseHeaderLexer, ElseIfHeaderLexer,
                              EndLexer, ExceptHeaderLexer, FinallyHeaderLexer,
                              ForHeaderLexer, IfHeaderLexer, ImplicitCommentLexer,
                              InlineIfHeaderLexer, InvalidSectionHeaderLexer,
                              KeywordCallLexer, KeywordSectionHeaderLexer,
                              KeywordSettingLexer, Lexer, ReturnLexer, SettingLexer,
                              SettingSectionHeaderLexer, SyntaxErrorLexer,
                              TaskSectionHeaderLexer, TestCaseSectionHeaderLexer,
                              TestCaseSettingLexer, TryHeaderLexer, VariableLexer,
                              VariableSectionHeaderLexer, WhileHeaderLexer)
from .tokens import StatementTokens, Token


class BlockLexer(Lexer, ABC):

    def __init__(self, ctx: LexingContext):
        super().__init__(ctx)
        self.lexers: 'list[Lexer]' = []

    def accepts_more(self, statement: StatementTokens) -> bool:
        return True

    def input(self, statement: StatementTokens):
        if self.lexers and self.lexers[-1].accepts_more(statement):
            lexer = self.lexers[-1]
        else:
            lexer = self.lexer_for(statement)
            self.lexers.append(lexer)
        lexer.input(statement)

    def lexer_for(self, statement: StatementTokens) -> Lexer:
        for cls in self.lexer_classes():
            lexer = cls(self.ctx)
            if lexer.handles(statement):
                return lexer
        raise TypeError(f"{type(self).__name__} does not have lexer for "
                        f"statement {statement}.")

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return ()

    def lex(self):
        for lexer in self.lexers:
            lexer.lex()

    def _lex_with_priority(self, priority: 'type[Lexer]'):
        for lexer in self.lexers:
            if isinstance(lexer, priority):
                lexer.lex()
        for lexer in self.lexers:
            if not isinstance(lexer, priority):
                lexer.lex()


class FileLexer(BlockLexer):

    def lex(self):
        self._lex_with_priority(priority=SettingSectionLexer)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (SettingSectionLexer, VariableSectionLexer,
                TestCaseSectionLexer, TaskSectionLexer,
                KeywordSectionLexer, CommentSectionLexer,
                InvalidSectionLexer, ImplicitCommentSectionLexer)


class SectionLexer(BlockLexer, ABC):
    ctx: FileContext

    def accepts_more(self, statement: StatementTokens) -> bool:
        return not statement[0].value.startswith('*')


class SettingSectionLexer(SectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return self.ctx.setting_section(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (SettingSectionHeaderLexer, SettingLexer)


class VariableSectionLexer(SectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return self.ctx.variable_section(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (VariableSectionHeaderLexer, VariableLexer)


class TestCaseSectionLexer(SectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return self.ctx.test_case_section(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (TestCaseSectionHeaderLexer, TestCaseLexer)


class TaskSectionLexer(SectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return self.ctx.task_section(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (TaskSectionHeaderLexer, TestCaseLexer)


class KeywordSectionLexer(SettingSectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return self.ctx.keyword_section(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (KeywordSectionHeaderLexer, KeywordLexer)


class CommentSectionLexer(SectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return self.ctx.comment_section(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (CommentSectionHeaderLexer, CommentLexer)


class ImplicitCommentSectionLexer(SectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return True

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (ImplicitCommentLexer,)


class InvalidSectionLexer(SectionLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return bool(statement and statement[0].value.startswith('*'))

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (InvalidSectionHeaderLexer, CommentLexer)


class TestOrKeywordLexer(BlockLexer, ABC):
    name_type: str
    _name_seen = False

    def accepts_more(self, statement: StatementTokens) -> bool:
        return not statement[0].value

    def input(self, statement: StatementTokens):
        self._handle_name_or_indentation(statement)
        if statement:
            super().input(statement)

    def _handle_name_or_indentation(self, statement: StatementTokens):
        if not self._name_seen:
            name_token = statement.pop(0)
            name_token.type = self.name_type
            if statement:
                name_token._add_eos_after = True
            self._name_seen = True
        else:
            while statement and not statement[0].value:
                statement.pop(0).type = None    # These tokens will be ignored


class TestCaseLexer(TestOrKeywordLexer):
    name_type = Token.TESTCASE_NAME

    def __init__(self, ctx: SuiteFileContext):
        super().__init__(ctx.test_case_context())

    def lex(self):
        self._lex_with_priority(priority=TestCaseSettingLexer)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (TestCaseSettingLexer, ForLexer, InlineIfLexer, IfLexer,
                TryLexer, WhileLexer, SyntaxErrorLexer, KeywordCallLexer)


class KeywordLexer(TestOrKeywordLexer):
    name_type = Token.KEYWORD_NAME

    def __init__(self, ctx: FileContext):
        super().__init__(ctx.keyword_context())

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (KeywordSettingLexer, ForLexer, InlineIfLexer, IfLexer, ReturnLexer,
                TryLexer, WhileLexer, SyntaxErrorLexer, KeywordCallLexer)


class NestedBlockLexer(BlockLexer, ABC):
    ctx: 'TestCaseContext|KeywordContext'

    def __init__(self, ctx: 'TestCaseContext|KeywordContext'):
        super().__init__(ctx)
        self._block_level = 0

    def accepts_more(self, statement: StatementTokens) -> bool:
        return self._block_level > 0

    def input(self, statement: StatementTokens):
        super().input(statement)
        lexer = self.lexers[-1]
        if isinstance(lexer, (ForHeaderLexer, IfHeaderLexer, TryHeaderLexer,
                              WhileHeaderLexer)):
            self._block_level += 1
        if isinstance(lexer, EndLexer):
            self._block_level -= 1


class ForLexer(NestedBlockLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return ForHeaderLexer(self.ctx).handles(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (ForHeaderLexer, InlineIfLexer, IfLexer, TryLexer, WhileLexer, EndLexer,
                ReturnLexer, ContinueLexer, BreakLexer, SyntaxErrorLexer, KeywordCallLexer)


class WhileLexer(NestedBlockLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return WhileHeaderLexer(self.ctx).handles(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (WhileHeaderLexer, ForLexer, InlineIfLexer, IfLexer, TryLexer, EndLexer,
                ReturnLexer, ContinueLexer, BreakLexer, SyntaxErrorLexer, KeywordCallLexer)


class TryLexer(NestedBlockLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return TryHeaderLexer(self.ctx).handles(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (TryHeaderLexer, ExceptHeaderLexer, ElseHeaderLexer, FinallyHeaderLexer,
                ForLexer, InlineIfLexer, IfLexer, WhileLexer, EndLexer, ReturnLexer,
                BreakLexer, ContinueLexer, SyntaxErrorLexer, KeywordCallLexer)


class IfLexer(NestedBlockLexer):

    def handles(self, statement: StatementTokens) -> bool:
        return IfHeaderLexer(self.ctx).handles(statement)

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (InlineIfLexer, IfHeaderLexer, ElseIfHeaderLexer, ElseHeaderLexer,
                ForLexer, TryLexer, WhileLexer, EndLexer, ReturnLexer, ContinueLexer,
                BreakLexer, SyntaxErrorLexer, KeywordCallLexer)


class InlineIfLexer(NestedBlockLexer):

    def handles(self, statement: StatementTokens) -> bool:
        if len(statement) <= 2:
            return False
        return InlineIfHeaderLexer(self.ctx).handles(statement)

    def accepts_more(self, statement: StatementTokens) -> bool:
        return False

    def lexer_classes(self) -> 'tuple[type[Lexer], ...]':
        return (InlineIfHeaderLexer, ElseIfHeaderLexer, ElseHeaderLexer,
                ReturnLexer, ContinueLexer, BreakLexer, KeywordCallLexer)

    def input(self, statement: StatementTokens):
        for part in self._split(statement):
            if part:
                super().input(part)

    def _split(self, statement: StatementTokens) -> 'Iterator[StatementTokens]':
        current = []
        expect_condition = False
        for token in statement:
            if expect_condition:
                if token is not statement[-1]:
                    token._add_eos_after = True
                current.append(token)
                yield current
                current = []
                expect_condition = False
            elif token.value == 'IF':
                current.append(token)
                expect_condition = True
            elif normalize_whitespace(token.value) == 'ELSE IF':
                token._add_eos_before = True
                yield current
                current = [token]
                expect_condition = True
            elif token.value == 'ELSE':
                token._add_eos_before = True
                if token is not statement[-1]:
                    token._add_eos_after = True
                yield current
                current = []
                yield [token]
            else:
                current.append(token)
        yield current



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.conf import RebotSettings
from robot.errors import DataError
from robot.model import ModelModifier
from robot.output import LOGGER
from robot.result import ExecutionResult, Result

from .jsmodelbuilders import JsModelBuilder
from .logreportwriters import LogWriter, ReportWriter
from .xunitwriter import XUnitWriter


class ResultWriter:
    """A class to create log, report, output XML and xUnit files.

    :param sources: Either one :class:`~robot.result.executionresult.Result`
        object, or one or more paths to existing output XML files.

    By default writes ``report.html`` and ``log.html``, but no output XML
    or xUnit files. Custom file names can be given and results disabled
    or enabled using ``settings`` or ``options`` passed to the
    :meth:`write_results` method. The latter is typically more convenient::

        writer = ResultWriter(result)
        writer.write_results(report='custom.html', log=None, xunit='xunit.xml')
    """

    def __init__(self, *sources):
        self._sources = sources

    def write_results(self, settings=None, **options):
        """Writes results based on the given ``settings``  or ``options``.

        :param settings: :class:`~robot.conf.settings.RebotSettings` object
            to configure result writing.
        :param options: Used to construct new
            :class:`~robot.conf.settings.RebotSettings` object if ``settings``
            are not given.
        """
        settings = settings or RebotSettings(options)
        results = Results(settings, *self._sources)
        if settings.output:
            self._write_output(results.result, settings.output)
        if settings.xunit:
            self._write_xunit(results.result, settings.xunit)
        if settings.log:
            config = dict(settings.log_config,
                          minLevel=results.js_result.min_level)
            self._write_log(results.js_result, settings.log, config)
        if settings.report:
            results.js_result.remove_data_not_needed_in_report()
            self._write_report(results.js_result, settings.report,
                               settings.report_config)
        return results.return_code

    def _write_output(self, result, path):
        self._write('Output', result.save, path)

    def _write_xunit(self, result, path):
        self._write('XUnit', XUnitWriter(result).write, path)

    def _write_log(self, js_result, path, config):
        self._write('Log', LogWriter(js_result).write, path, config)

    def _write_report(self, js_result, path, config):
        self._write('Report', ReportWriter(js_result).write, path, config)

    def _write(self, name, writer, path, *args):
        try:
            writer(path, *args)
        except DataError as err:
            LOGGER.error(err.message)
        else:
            LOGGER.output_file(name, path)


class Results:

    def __init__(self, settings, *sources):
        self._settings = settings
        self._sources = sources
        if len(sources) == 1 and isinstance(sources[0], Result):
            self._result = sources[0]
            self._prune = False
            self.return_code = self._result.return_code
        else:
            self._result = None
            self._prune = True
            self.return_code = -1
        self._js_result = None

    @property
    def result(self):
        if self._result is None:
            include_keywords = bool(self._settings.log or self._settings.output)
            flattened = self._settings.flatten_keywords
            self._result = ExecutionResult(include_keywords=include_keywords,
                                           flattened_keywords=flattened,
                                           merge=self._settings.merge,
                                           rpa=self._settings.rpa,
                                           *self._sources)
            if self._settings.rpa is None:
                self._settings.rpa = self._result.rpa
            modifier = ModelModifier(self._settings.pre_rebot_modifiers,
                                     self._settings.process_empty_suite,
                                     LOGGER)
            self._result.suite.visit(modifier)
            self._result.configure(self._settings.status_rc,
                                   self._settings.suite_config,
                                   self._settings.statistics_config)
            self.return_code = self._result.return_code
        return self._result

    @property
    def js_result(self):
        if self._js_result is None:
            builder = JsModelBuilder(log_path=self._settings.log,
                                     split_log=self._settings.split_log,
                                     expand_keywords=self._settings.expand_keywords,
                                     prune_input_to_save_memory=self._prune)
            self._js_result = builder.build_from(self.result)
            if self._prune:
                self._result = None
        return self._js_result



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import fnmatch
import os.path
import re
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Iterable, Iterator, Sequence

from robot.errors import DataError
from robot.output import LOGGER
from robot.utils import get_error_message


class SuiteStructure(ABC):
    source: 'Path|None'
    init_file: 'Path|None'
    children: 'list[SuiteStructure]|None'

    def __init__(self, extensions: 'ValidExtensions', source: 'Path|None',
                 init_file: 'Path|None' = None,
                 children: 'Sequence[SuiteStructure]|None' = None):
        self._extensions = extensions
        self.source = source
        self.init_file = init_file
        self.children = list(children) if children is not None else None

    @property
    def extension(self) -> 'str|None':
        source = self._get_source_file()
        return self._extensions.get_extension(source) if source else None

    @abstractmethod
    def _get_source_file(self) -> 'Path|None':
        raise NotImplementedError

    @abstractmethod
    def visit(self, visitor: 'SuiteStructureVisitor'):
        raise NotImplementedError


class SuiteFile(SuiteStructure):
    source: Path

    def __init__(self, extensions: 'ValidExtensions', source: Path):
        super().__init__(extensions, source)

    def _get_source_file(self) -> Path:
        return self.source

    def visit(self, visitor: 'SuiteStructureVisitor'):
        visitor.visit_file(self)


class SuiteDirectory(SuiteStructure):
    children: 'list[SuiteStructure]'

    def __init__(self, extensions: 'ValidExtensions', source: 'Path|None' = None,
                 init_file: 'Path|None' = None,
                 children: Sequence[SuiteStructure] = ()):
        super().__init__(extensions, source, init_file, children)

    def _get_source_file(self) -> 'Path|None':
        return self.init_file

    @property
    def is_multi_source(self) -> bool:
        return self.source is None

    def add(self, child: 'SuiteStructure'):
        self.children.append(child)

    def visit(self, visitor: 'SuiteStructureVisitor'):
        visitor.visit_directory(self)


class SuiteStructureVisitor:

    def visit_file(self, structure: SuiteFile):
        pass

    def visit_directory(self, structure: SuiteDirectory):
        self.start_directory(structure)
        for child in structure.children:
            child.visit(self)
        self.end_directory(structure)

    def start_directory(self, structure: SuiteDirectory):
        pass

    def end_directory(self, structure: SuiteDirectory):
        pass


class SuiteStructureBuilder:
    ignored_prefixes = ('_', '.')
    ignored_dirs = ('CVS',)

    def __init__(self, extensions: Sequence[str] = ('.robot', '.rbt', '.robot.rst'),
                 included_files: Sequence[str] = ()):
        self.extensions = ValidExtensions(extensions, included_files)
        self.included_files = IncludedFiles(included_files)

    def build(self, *paths: Path) -> SuiteStructure:
        if len(paths) == 1:
            return self._build(paths[0])
        return self._build_multi_source(paths)

    def _build(self, path: Path) -> SuiteStructure:
        if path.is_file():
            return SuiteFile(self.extensions, path)
        return self._build_directory(path)

    def _build_directory(self, path: Path) -> SuiteStructure:
        structure = SuiteDirectory(self.extensions, path)
        for item in self._list_dir(path):
            if self._is_init_file(item):
                if structure.init_file:
                    # TODO: This error should fail parsing for good.
                    LOGGER.error(f"Ignoring second test suite init file '{item}'.")
                else:
                    structure.init_file = item
            elif self._is_included(item):
                structure.add(self._build(item))
            else:
                LOGGER.info(f"Ignoring file or directory '{item}'.")
        return structure

    def _list_dir(self, path: Path) -> 'list[Path]':
        try:
            return sorted(path.iterdir(), key=lambda p: p.name.lower())
        except OSError:
            raise DataError(f"Reading directory '{path}' failed: {get_error_message()}")

    def _is_init_file(self, path: Path) -> bool:
        return (path.stem.lower() == '__init__'
                and self.extensions.match(path)
                and path.is_file())

    def _is_included(self, path: Path) -> bool:
        if path.name.startswith(self.ignored_prefixes):
            return False
        if path.is_dir():
            return path.name not in self.ignored_dirs
        if not path.is_file():
            return False
        if not self.extensions.match(path):
            return False
        return self.included_files.match(path)

    def _build_multi_source(self, paths: Iterable[Path]) -> SuiteStructure:
        structure = SuiteDirectory(self.extensions)
        for path in paths:
            if self._is_init_file(path):
                if structure.init_file:
                    raise DataError("Multiple init files not allowed.")
                structure.init_file = path
            else:
                structure.add(self._build(path))
        return structure


class ValidExtensions:

    def __init__(self, extensions: Sequence[str],
                 included_files: Sequence[str] = ()):
        self.extensions = {ext.lstrip('.').lower() for ext in extensions}
        for pattern in included_files:
            ext = os.path.splitext(pattern)[1]
            if ext:
                self.extensions.add(ext.lstrip('.').lower())

    def match(self, path: Path) -> bool:
        for ext in self._extensions_from(path):
            if ext in self.extensions:
                return True
        return False

    def get_extension(self, path: Path) -> str:
        for ext in self._extensions_from(path):
            if ext in self.extensions:
                return ext
        return path.suffix.lower()[1:]

    def _extensions_from(self, path: Path) -> Iterator[str]:
        suffixes = path.suffixes
        while suffixes:
            yield ''.join(suffixes).lower()[1:]
            suffixes.pop(0)


class IncludedFiles:

    def __init__(self, patterns: 'Sequence[str|Path]' = ()):
        self.patterns = [self._compile(i) for i in patterns]

    def _compile(self, pattern: 'str|Path') -> 're.Pattern':
        pattern = self._dir_to_recursive(self._path_to_abs(self._normalize(pattern)))
        # Handle recursive glob patterns.
        parts = [self._translate(p) for p in pattern.split('**')]
        return re.compile('.*'.join(parts), re.IGNORECASE)

    def _normalize(self, pattern: 'str|Path') -> str:
        if isinstance(pattern, Path):
            pattern = str(pattern)
        return os.path.normpath(pattern).replace('\\', '/')

    def _path_to_abs(self, pattern: str) -> str:
        if '/' in pattern or '.' not in pattern or os.path.exists(pattern):
            pattern = os.path.abspath(pattern).replace('\\', '/')
        return pattern

    def _dir_to_recursive(self, pattern: str) -> str:
        if '.' not in os.path.basename(pattern) or os.path.isdir(pattern):
            pattern += '/**'
        return pattern

    def _translate(self, glob_pattern: str) -> str:
        # `fnmatch.translate` returns pattern in format `(?s:<pattern>)\Z` but we want
        # only the `<pattern>` part. This is a bit risky because the format may change
        # in future Python versions, but we have tests and ought to notice that.
        re_pattern = fnmatch.translate(glob_pattern)[4:-3]
        # Unlike `fnmatch`, we want `*` to match only a single path segment.
        return re_pattern.replace('.*', '[^/]*')

    def match(self, path: Path) -> bool:
        if not self.patterns:
            return True
        return self._match(path.name) or self._match(str(path))

    def _match(self, path: str) -> bool:
        path = self._normalize(path)
        return any(p.fullmatch(path) for p in self.patterns)



print("import: 'robot'")
import robot

print("import: 'robot.api'")
import robot.api

print("import: 'robot.conf'")
import robot.conf

print("import: 'robot.htmldata'")
import robot.htmldata

print("import: 'robot.libdocpkg'")
import robot.libdocpkg

print("import: 'robot.libraries'")
import robot.libraries

print("import: 'robot.model'")
import robot.model

print("import: 'robot.output'")
import robot.output

print("import: 'robot.output.console'")
import robot.output.console

print("import: 'robot.parsing'")
import robot.parsing

print("import: 'robot.reporting'")
import robot.reporting

print("import: 'robot.result'")
import robot.result

print("import: 'robot.running'")
import robot.running

print("import: 'robot.running.arguments'")
import robot.running.arguments

print("import: 'robot.running.timeouts'")
import robot.running.timeouts

print("import: 'robot.utils'")
import robot.utils

print("import: 'robot.variables'")
import robot.variables




#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

from robot.errors import DataError, VariableError
from robot.utils import (get_env_var, get_env_vars, get_error_message, normalize,
                         NormalizedDict)

from .evaluation import evaluate_expression
from .notfound import variable_not_found
from .search import search_variable, VariableMatch


NOT_FOUND = object()


class VariableFinder:

    def __init__(self, variables):
        self._finders = (StoredFinder(variables.store),
                         NumberFinder(),
                         EmptyFinder(),
                         InlinePythonFinder(variables),
                         EnvironmentFinder(),
                         ExtendedFinder(self))
        self._store = variables.store

    def find(self, variable):
        match = self._get_match(variable)
        name = match.name
        for finder in self._finders:
            if match.identifier in finder.identifiers:
                result = finder.find(name)
                if result is not NOT_FOUND:
                    return result
        variable_not_found(name, self._store.data)

    def _get_match(self, variable):
        if isinstance(variable, VariableMatch):
            return variable
        match = search_variable(variable)
        if not match.is_variable() or match.items:
            raise DataError("Invalid variable name '%s'." % variable)
        return match


class StoredFinder:
    identifiers = '$@&'

    def __init__(self, store):
        self._store = store

    def find(self, name):
        return self._store.get(name, NOT_FOUND)


class NumberFinder:
    identifiers = '$'

    def find(self, name):
        number = normalize(name)[2:-1]
        for converter in self._get_int, float:
            try:
                return converter(number)
            except ValueError:
                pass
        return NOT_FOUND

    def _get_int(self, number):
        bases = {'0b': 2, '0o': 8, '0x': 16}
        if number.startswith(tuple(bases)):
            return int(number[2:], bases[number[:2]])
        return int(number)


class EmptyFinder:
    identifiers = '$@&'
    empty = NormalizedDict({'${EMPTY}': '', '@{EMPTY}': (), '&{EMPTY}': {}}, ignore='_')

    def find(self, name):
        return self.empty.get(name, NOT_FOUND)


class InlinePythonFinder:
    identifiers = '$@&'

    def __init__(self, variables):
        self._variables = variables

    def find(self, name):
        base = name[2:-1]
        if not base or base[0] != '{' or base[-1] != '}':
            return NOT_FOUND
        try:
            return evaluate_expression(base[1:-1].strip(), self._variables)
        except DataError as err:
            raise VariableError("Resolving variable '%s' failed: %s" % (name, err))


class ExtendedFinder:
    identifiers = '$@&'
    _match_extended = re.compile(r'''
        (.+?)          # base name (group 1)
        ([^\s\w].+)    # extended part (group 2)
    ''', re.UNICODE|re.VERBOSE).match

    def __init__(self, finder):
        self._find_variable = finder.find

    def find(self, name):
        match = self._match_extended(name[2:-1])
        if match is None:
            return NOT_FOUND
        base_name, extended = match.groups()
        try:
            variable = self._find_variable('${%s}' % base_name)
        except DataError as err:
            raise VariableError("Resolving variable '%s' failed: %s"
                                % (name, err.message))
        try:
            return eval('_BASE_VAR_' + extended, {'_BASE_VAR_': variable})
        except:
            raise VariableError("Resolving variable '%s' failed: %s"
                                % (name, get_error_message()))


class EnvironmentFinder:
    identifiers = '%'

    def find(self, name):
        var_name, has_default, default_value = name[2:-1].partition('=')
        value = get_env_var(var_name)
        if value is not None:
            return value
        if has_default:
            return default_value
        variable_not_found(name, get_env_vars(),
                           "Environment variable '%s' not found." % name)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from itertools import chain

from robot.errors import (BreakLoop, ContinueLoop, DataError, ExecutionFailed,
                          ExecutionPassed, ExecutionStatus, PassExecution,
                          ReturnFromKeyword, UserKeywordExecutionFailed, VariableError)
from robot.result import Keyword as KeywordResult
from robot.utils import DotDict, getshortdoc, prepr, split_tags_from_doc
from robot.variables import is_list_variable, VariableAssignment

from .arguments import DefaultValue
from .bodyrunner import BodyRunner, KeywordRunner
from .statusreporter import StatusReporter
from .timeouts import KeywordTimeout


class UserKeywordRunner:

    def __init__(self, handler, name=None):
        self._handler = handler
        self.name = name or handler.name
        self.pre_run_messages = ()

    @property
    def longname(self):
        libname = self._handler.libname
        return f'{libname}.{self.name}' if libname else self.name

    @property
    def libname(self):
        return self._handler.libname

    @property
    def tags(self):
        return self._handler.tags

    @property
    def source(self):
        return self._handler.source

    @property
    def arguments(self):
        """:rtype: :py:class:`robot.running.arguments.ArgumentSpec`"""
        return self._handler.arguments

    def run(self, kw, context, run=True):
        assignment = VariableAssignment(kw.assign)
        result = self._get_result(kw, assignment, context.variables)
        with StatusReporter(kw, result, context, run):
            if self._handler.private:
                context.warn_on_invalid_private_call(self._handler)
            with assignment.assigner(context) as assigner:
                if run:
                    return_value = self._run(context, kw.args, result)
                    assigner.assign(return_value)
                    return return_value

    def _get_result(self, kw, assignment, variables):
        handler = self._handler
        doc = variables.replace_string(handler.doc, ignore_errors=True)
        doc, tags = split_tags_from_doc(doc)
        tags = variables.replace_list(handler.tags, ignore_errors=True) + tags
        return KeywordResult(kwname=self.name,
                             libname=handler.libname,
                             doc=getshortdoc(doc),
                             args=kw.args,
                             assign=tuple(assignment),
                             tags=tags,
                             type=kw.type)

    def _run(self, context, args, result):
        if self.pre_run_messages:
            for message in self.pre_run_messages:
                context.output.message(message)
        variables = context.variables
        args = self._resolve_arguments(args, variables)
        with context.user_keyword(self._handler):
            self._set_arguments(args, context)
            timeout = self._get_timeout(variables)
            if timeout is not None:
                result.timeout = str(timeout)
            with context.timeout(timeout):
                exception, return_ = self._execute(context)
                if exception and not exception.can_continue(context):
                    raise exception
                return_value = self._get_return_value(variables, return_)
                if exception:
                    exception.return_value = return_value
                    raise exception
                return return_value

    def _get_timeout(self, variables=None):
        timeout = self._handler.timeout
        return KeywordTimeout(timeout, variables) if timeout else None

    def _resolve_arguments(self, arguments, variables=None):
        return self.arguments.resolve(arguments, variables)

    def _set_arguments(self, arguments, context):
        positional, named = arguments
        variables = context.variables
        args, kwargs = self.arguments.map(positional, named,
                                          replace_defaults=False)
        self._set_variables(args, kwargs, variables)
        context.output.trace(lambda: self._trace_log_args_message(variables),
                             write_if_flat=False)

    def _set_variables(self, positional, kwargs, variables):
        spec = self.arguments
        args, varargs = self._split_args_and_varargs(positional)
        kwonly, kwargs = self._split_kwonly_and_kwargs(kwargs)
        for name, value in chain(zip(spec.positional, args), kwonly):
            if isinstance(value, DefaultValue):
                value = value.resolve(variables)
            variables[f'${{{name}}}'] = value
        if spec.var_positional:
            variables[f'@{{{spec.var_positional}}}'] = varargs
        if spec.var_named:
            variables[f'&{{{spec.var_named}}}'] = DotDict(kwargs)

    def _split_args_and_varargs(self, args):
        if not self.arguments.var_positional:
            return args, []
        positional = len(self.arguments.positional)
        return args[:positional], args[positional:]

    def _split_kwonly_and_kwargs(self, all_kwargs):
        kwonly = []
        kwargs = []
        for name, value in all_kwargs:
            target = kwonly if name in self.arguments.named_only else kwargs
            target.append((name, value))
        return kwonly, kwargs

    def _trace_log_args_message(self, variables):
        return self._format_trace_log_args_message(
            self._format_args_for_trace_logging(), variables)

    def _format_args_for_trace_logging(self):
        args = [f'${{{arg}}}' for arg in self.arguments.positional]
        if self.arguments.var_positional:
            args.append(f'@{{{self.arguments.var_positional}}}')
        if self.arguments.var_named:
            args.append(f'&{{{self.arguments.var_named}}}')
        return args

    def _format_trace_log_args_message(self, args, variables):
        args = ' | '.join(f'{name}={prepr(variables[name])}' for name in args)
        return f'Arguments: [ {args} ]'

    def _execute(self, context):
        handler = self._handler
        if context.dry_run and handler.tags.robot('no-dry-run'):
            return None, None
        error = return_ = pass_ = None
        try:
            BodyRunner(context).run(handler.body)
        except ReturnFromKeyword as exception:
            return_ = exception
            error = exception.earlier_failures
        except (BreakLoop, ContinueLoop) as exception:
            pass_ = exception
        except ExecutionPassed as exception:
            pass_ = exception
            error = exception.earlier_failures
            if error:
                error.continue_on_failure = False
        except ExecutionFailed as exception:
            error = exception
        if handler.teardown:
            with context.keyword_teardown(error):
                td_error = self._run_teardown(handler.teardown, context)
        else:
            td_error = None
        if error or td_error:
            error = UserKeywordExecutionFailed(error, td_error)
        return error or pass_, return_

    def _get_return_value(self, variables, return_):
        ret = self._handler.return_value if not return_ else return_.return_value
        if not ret:
            return None
        contains_list_var = any(is_list_variable(item) for item in ret)
        try:
            ret = variables.replace_list(ret)
        except DataError as err:
            raise VariableError(f'Replacing variables from keyword return '
                                f'value failed: {err}')
        if len(ret) != 1 or contains_list_var:
            return ret
        return ret[0]

    def _run_teardown(self, teardown, context):
        try:
            name = context.variables.replace_string(teardown.name)
        except DataError as err:
            if context.dry_run:
                return None
            return ExecutionFailed(err.message, syntax=True)
        if name.upper() in ('', 'NONE'):
            return None
        try:
            KeywordRunner(context).run(teardown, name)
        except PassExecution:
            return None
        except ExecutionStatus as err:
            return err
        return None

    def dry_run(self, kw, context):
        assignment = VariableAssignment(kw.assign)
        result = self._get_result(kw, assignment, context.variables)
        with StatusReporter(kw, result, context):
            assignment.validate_assignment()
            self._dry_run(context, kw.args, result)

    def _dry_run(self, context, args, result):
        if self.pre_run_messages:
            for message in self.pre_run_messages:
                context.output.message(message)
        self._resolve_arguments(args)
        with context.user_keyword(self._handler):
            timeout = self._get_timeout()
            if timeout:
                result.timeout = str(timeout)
            error, _ = self._execute(context)
            if error:
                raise error


class EmbeddedArgumentsRunner(UserKeywordRunner):

    def __init__(self, handler, name):
        super().__init__(handler, name)
        self.embedded_args = handler.embedded.match(name).groups()

    def _resolve_arguments(self, args, variables=None):
        self.arguments.resolve(args, variables)
        if variables:
            embedded = [variables.replace_scalar(e) for e in self.embedded_args]
            self.embedded_args = self._handler.embedded.map(embedded)
        return super()._resolve_arguments(args, variables)

    def _set_arguments(self, args, context):
        variables = context.variables
        for name, value in self.embedded_args:
            variables[f'${{{name}}}'] = value
        super()._set_arguments(args, context)
        context.output.trace(lambda: self._trace_log_args_message(variables),
                             write_if_flat=False)

    def _trace_log_args_message(self, variables):
        args = [f'${{{arg}}}' for arg in self._handler.embedded.args]
        args += self._format_args_for_trace_logging()
        return self._format_trace_log_args_message(args, variables)

    def _get_result(self, kw, assignment, variables):
        result = UserKeywordRunner._get_result(self, kw, assignment, variables)
        result.sourcename = self._handler.name
        return result



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod

from robot.conf import Languages
from robot.utils import normalize, normalize_whitespace, RecommendationFinder

from .tokens import StatementTokens, Token


class Settings(ABC):
    names: 'tuple[str, ...]' = ()
    aliases: 'dict[str, str]' = {}
    multi_use = (
        'Metadata',
        'Library',
        'Resource',
        'Variables'
    )
    single_value = (
        'Resource',
        'Test Timeout',
        'Test Template',
        'Timeout',
        'Template',
        'Name'
    )
    name_and_arguments = (
        'Metadata',
        'Suite Setup',
        'Suite Teardown',
        'Test Setup',
        'Test Teardown',
        'Test Template',
        'Setup',
        'Teardown',
        'Template',
        'Resource',
        'Variables'
    )
    name_arguments_and_with_name = (
        'Library',
    )

    def __init__(self, languages: Languages):
        self.settings: 'dict[str, list[Token]|None]' = {n: None for n in self.names}
        self.languages = languages

    def lex(self, statement: StatementTokens):
        orig = self._format_name(statement[0].value)
        name = normalize_whitespace(orig).title()
        name = self.languages.settings.get(name, name)
        if name in self.aliases:
            name = self.aliases[name]
        try:
            self._validate(orig, name, statement)
        except ValueError as err:
            self._lex_error(statement, err.args[0])
        else:
            self._lex_setting(statement, name)

    def _format_name(self, name: str) -> str:
        return name

    def _validate(self, orig: str, name: str, statement: StatementTokens):
        if name not in self.settings:
            message = self._get_non_existing_setting_message(orig, name)
            raise ValueError(message)
        if self.settings[name] is not None and name not in self.multi_use:
            raise ValueError(f"Setting '{orig}' is allowed only once. "
                             f"Only the first value is used.")
        if name in self.single_value and len(statement) > 2:
            raise ValueError(f"Setting '{orig}' accepts only one value, "
                             f"got {len(statement)-1}.")

    def _get_non_existing_setting_message(self, name: str, normalized: str) -> str:
        if self._is_valid_somewhere(normalized, Settings.__subclasses__()):
            return self._not_valid_here(name)
        return RecommendationFinder(normalize).find_and_format(
            name=normalized,
            candidates=tuple(self.settings) + tuple(self.aliases),
            message=f"Non-existing setting '{name}'."
        )

    def _is_valid_somewhere(self, name: str, classes: 'list[type[Settings]]') -> bool:
        for cls in classes:
            if (name in cls.names or name in cls.aliases
                    or self._is_valid_somewhere(name, cls.__subclasses__())):
                return True
        return False

    @abstractmethod
    def _not_valid_here(self, name: str) -> str:
        raise NotImplementedError

    def _lex_error(self, statement: StatementTokens, error: str):
        statement[0].set_error(error)
        for token in statement[1:]:
            token.type = Token.COMMENT

    def _lex_setting(self, statement: StatementTokens, name: str):
        # TODO: Change token type from 'FORCE TAGS' to 'TEST TAGS' in RF 7.0.
        statement[0].type = {'Test Tags': Token.FORCE_TAGS,
                             'Name': Token.SUITE_NAME}.get(name, name.upper())
        self.settings[name] = values = statement[1:]
        if name in self.name_and_arguments:
            self._lex_name_and_arguments(values)
        elif name in self.name_arguments_and_with_name:
            self._lex_name_arguments_and_with_name(values)
        else:
            self._lex_arguments(values)

    def _lex_name_and_arguments(self, tokens: StatementTokens):
        if tokens:
            tokens[0].type = Token.NAME
            self._lex_arguments(tokens[1:])

    def _lex_name_arguments_and_with_name(self, tokens: StatementTokens):
        self._lex_name_and_arguments(tokens)
        if len(tokens) > 1 and \
                normalize_whitespace(tokens[-2].value) in ('WITH NAME', 'AS'):
            tokens[-2].type = Token.WITH_NAME
            tokens[-1].type = Token.NAME

    def _lex_arguments(self, tokens: StatementTokens):
        for token in tokens:
            token.type = Token.ARGUMENT


class FileSettings(Settings, ABC):
    pass


class SuiteFileSettings(FileSettings):
    names = (
        'Documentation',
        'Metadata',
        'Name',
        'Suite Setup',
        'Suite Teardown',
        'Test Setup',
        'Test Teardown',
        'Test Template',
        'Test Timeout',
        'Test Tags',
        'Default Tags',
        'Keyword Tags',
        'Library',
        'Resource',
        'Variables'
    )
    aliases = {
        'Force Tags': 'Test Tags',
        'Task Tags': 'Test Tags',
        'Task Setup': 'Test Setup',
        'Task Teardown': 'Test Teardown',
        'Task Template': 'Test Template',
        'Task Timeout': 'Test Timeout',
    }

    def _not_valid_here(self, name: str) -> str:
        return f"Setting '{name}' is not allowed in suite file."


class InitFileSettings(FileSettings):
    names = (
        'Documentation',
        'Metadata',
        'Name',
        'Suite Setup',
        'Suite Teardown',
        'Test Setup',
        'Test Teardown',
        'Test Timeout',
        'Test Tags',
        'Keyword Tags',
        'Library',
        'Resource',
        'Variables'
    )
    aliases = {
        'Force Tags': 'Test Tags',
        'Task Tags': 'Test Tags',
        'Task Setup': 'Test Setup',
        'Task Teardown': 'Test Teardown',
        'Task Timeout': 'Test Timeout',
    }

    def _not_valid_here(self, name: str) -> str:
        return f"Setting '{name}' is not allowed in suite initialization file."


class ResourceFileSettings(FileSettings):
    names = (
        'Documentation',
        'Keyword Tags',
        'Library',
        'Resource',
        'Variables'
    )

    def _not_valid_here(self, name: str) -> str:
        return f"Setting '{name}' is not allowed in resource file."


class TestCaseSettings(Settings):
    names = (
        'Documentation',
        'Tags',
        'Setup',
        'Teardown',
        'Template',
        'Timeout'
    )

    def __init__(self, parent: SuiteFileSettings):
        super().__init__(parent.languages)
        self.parent = parent

    def _format_name(self, name: str) -> str:
        return name[1:-1].strip()

    @property
    def template_set(self) -> bool:
        template = self.settings['Template']
        if self._has_disabling_value(template):
            return False
        parent_template = self.parent.settings['Test Template']
        return self._has_value(template) or self._has_value(parent_template)

    def _has_disabling_value(self, setting: 'StatementTokens|None') -> bool:
        if setting is None:
            return False
        return setting == [] or setting[0].value.upper() == 'NONE'

    def _has_value(self, setting: 'StatementTokens|None') -> bool:
        return bool(setting and setting[0].value)

    def _not_valid_here(self, name: str) -> str:
        return f"Setting '{name}' is not allowed with tests or tasks."


class KeywordSettings(Settings):
    names = (
        'Documentation',
        'Arguments',
        'Teardown',
        'Timeout',
        'Tags',
        'Return'
    )

    def __init__(self, parent: FileSettings):
        super().__init__(parent.languages)
        self.parent = parent

    def _format_name(self, name: str) -> str:
        return name[1:-1].strip()

    def _not_valid_here(self, name: str) -> str:
        return f"Setting '{name}' is not allowed with user keywords."



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.htmldata import HtmlFileWriter, ModelWriter, LIBDOC


class LibdocHtmlWriter:

    def __init__(self, theme=None):
        self.theme = theme

    def write(self, libdoc, output):
        model_writer = LibdocModelWriter(output, libdoc, self.theme)
        HtmlFileWriter(output, model_writer).write(LIBDOC)


class LibdocModelWriter(ModelWriter):

    def __init__(self, output, libdoc, theme=None):
        self.output = output
        self.libdoc = libdoc
        self.theme = theme

    def write(self, line):
        data = self.libdoc.to_json(include_private=False, theme=self.theme)
        self.output.write(f'<script type="text/javascript">\n'
                          f'libdoc = {data}\n'
                          f'</script>\n')



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from operator import eq, lt, le, gt, ge

from .robottypes import type_name


class Sortable:
    """Base class for sorting based self._sort_key"""

    _sort_key = NotImplemented

    def __test(self, operator, other, require_sortable=True):
        if isinstance(other, Sortable):
            return operator(self._sort_key, other._sort_key)
        if not require_sortable:
            return False
        raise TypeError("Cannot sort '%s' and '%s'."
                        % (type_name(self), type_name(other)))

    def __eq__(self, other):
        return self.__test(eq, other, require_sortable=False)

    def __lt__(self, other):
        return self.__test(lt, other)

    def __le__(self, other):
        return self.__test(le, other)

    def __gt__(self, other):
        return self.__test(gt, other)

    def __ge__(self, other):
        return self.__test(ge, other)

    def __hash__(self):
        return hash(self._sort_key)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

from robot.errors import DataError
from robot.utils import get_error_message, is_string
from robot.variables import VariableIterator


ENABLE_STRICT_ARGUMENT_VALIDATION = False


class EmbeddedArguments:

    def __init__(self, name=None, args=(), custom_patterns=None):
        self.name = name
        self.args = args
        self.custom_patterns = custom_patterns

    @classmethod
    def from_name(cls, name):
        return EmbeddedArgumentParser().parse(name) if '${' in name else cls()

    def match(self, name):
        return self.name.match(name)

    def map(self, values):
        self.validate(values)
        return list(zip(self.args, values))

    def validate(self, values):
        # Validating that embedded args match custom regexps also if args are
        # given as variables was initially implemented in RF 6.0. It needed
        # to be reverted due to backwards incompatibility reasons but the plan
        # is to enable it again in RF 7.0:
        # https://github.com/robotframework/robotframework/issues/4069
        #
        # TODO: Emit deprecation warnings if patterns don't match in RF 6.1:
        # https://github.com/robotframework/robotframework/issues/4524
        #
        # Because the plan is to add validation back, the code was not removed
        # but the `ENABLE_STRICT_ARGUMENT_VALIDATION` guard was added instead.
        # Enabling validation requires only removing the following two lines
        # (along with this comment). If someone wants to enable strict validation
        # already now, they set `ENABLE_STRICT_ARGUMENT_VALIDATION` to True
        # before running tests.
        if not ENABLE_STRICT_ARGUMENT_VALIDATION:
            return
        if not self.custom_patterns:
            return
        for arg, value in zip(self.args, values):
            if arg in self.custom_patterns and is_string(value):
                pattern = self.custom_patterns[arg]
                if not re.fullmatch(pattern, value):
                    raise ValueError(f"Embedded argument '{arg}' got value '{value}' "
                                     f"that does not match custom pattern '{pattern}'.")

    def __bool__(self):
        return self.name is not None


class EmbeddedArgumentParser:
    _regexp_extension = re.compile(r'(?<!\\)\(\?.+\)')
    _regexp_group_start = re.compile(r'(?<!\\)\((.*?)\)')
    _escaped_curly = re.compile(r'(\\+)([{}])')
    _regexp_group_escape = r'(?:\1)'
    _default_pattern = '.*?'
    _variable_pattern = r'\$\{[^\}]+\}'

    def parse(self, string):
        args = []
        custom_patterns = {}
        name_regexp = ['^']
        for before, variable, string in VariableIterator(string, identifiers='$'):
            name, pattern, custom = self._get_name_and_pattern(variable[2:-1])
            args.append(name)
            if custom:
                custom_patterns[name] = pattern
                pattern = self._format_custom_regexp(pattern)
            name_regexp.extend([re.escape(before), f'({pattern})'])
        name_regexp.extend([re.escape(string), '$'])
        name = self._compile_regexp(name_regexp) if args else None
        return EmbeddedArguments(name, args, custom_patterns or None)

    def _get_name_and_pattern(self, name):
        if ':' in name:
            name, pattern = name.split(':', 1)
            custom = True
        else:
            pattern = self._default_pattern
            custom = False
        return name, pattern, custom

    def _format_custom_regexp(self, pattern):
        for formatter in (self._regexp_extensions_are_not_allowed,
                          self._make_groups_non_capturing,
                          self._unescape_curly_braces,
                          self._escape_escapes,
                          self._add_automatic_variable_pattern):
            pattern = formatter(pattern)
        return pattern

    def _regexp_extensions_are_not_allowed(self, pattern):
        if self._regexp_extension.search(pattern):
            raise DataError('Regexp extensions are not allowed in embedded arguments.')
        return pattern

    def _make_groups_non_capturing(self, pattern):
        return self._regexp_group_start.sub(self._regexp_group_escape, pattern)

    def _unescape_curly_braces(self, pattern):
        # Users must escape possible lone curly braces in patters (e.g. `${x:\{}`)
        # or otherwise the variable syntax is invalid.
        def unescape(match):
            backslashes = len(match.group(1))
            return '\\' * (backslashes // 2 * 2) + match.group(2)
        return self._escaped_curly.sub(unescape, pattern)

    def _escape_escapes(self, pattern):
        # When keywords are matched, embedded arguments have not yet been
        # resolved which means possible escapes are still doubled. We thus
        # need to double them in the pattern as well.
        return pattern.replace(r'\\', r'\\\\')

    def _add_automatic_variable_pattern(self, pattern):
        return f'{pattern}|{self._variable_pattern}'

    def _compile_regexp(self, pattern):
        try:
            return re.compile(''.join(pattern), re.IGNORECASE)
        except Exception:
            raise DataError("Compiling embedded arguments regexp failed: %s"
                            % get_error_message())



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Various generic utility functions and classes.

Utilities are mainly for internal usage, but external libraries and tools
may find some of them useful. Utilities are generally stable, but absolute
backwards compatibility between major versions is not guaranteed.

All utilities are exposed via the :mod:`robot.utils` package, and should be
used either like::

    from robot import utils

    assert utils.Matcher('H?llo').match('Hillo')

or::

    from robot.utils import Matcher

    assert Matcher('H?llo').match('Hillo')
"""

from .argumentparser import ArgumentParser, cmdline2list
from .application import Application
from .compress import compress_text
from .connectioncache import ConnectionCache
from .dotdict import DotDict
from .encoding import (CONSOLE_ENCODING, SYSTEM_ENCODING, console_decode,
                       console_encode, system_decode, system_encode)
from .error import (get_error_message, get_error_details, ErrorDetails)
from .escaping import escape, glob_escape, unescape, split_from_equals
from .etreewrapper import ET, ETSource
from .filereader import FileReader, Source
from .frange import frange
from .markuputils import html_format, html_escape, xml_escape, attribute_escape
from .markupwriters import HtmlWriter, XmlWriter, NullMarkupWriter
from .importer import Importer
from .match import eq, Matcher, MultiMatcher
from .misc import (classproperty, isatty, parse_re_flags, plural_or_not,
                   printable_name, seq2str, seq2str2, test_or_task)
from .normalizing import normalize, normalize_whitespace, NormalizedDict
from .platform import PY_VERSION, PYPY, UNIXY, WINDOWS, RERAISED_EXCEPTIONS
from .recommendations import RecommendationFinder
from .robotenv import get_env_var, set_env_var, del_env_var, get_env_vars
from .robotinspect import is_init
from .robotio import binary_file_writer, create_destination_directory, file_writer
from .robotpath import abspath, find_file, get_link_path, normpath
from .robottime import (elapsed_time_to_string, format_time, get_elapsed_time,
                        get_time, get_timestamp, secs_to_timestamp,
                        secs_to_timestr, timestamp_to_secs, timestr_to_secs,
                        parse_time)
from .robottypes import (has_args, is_bytes, is_dict_like, is_falsy, is_integer,
                         is_list_like, is_number, is_pathlike, is_string, is_truthy,
                         is_union, type_name, type_repr, typeddict_types)
from .setter import setter, SetterAwareType
from .sortable import Sortable
from .text import (cut_assign_value, cut_long_message, format_assign_message,
                   get_console_length, getdoc, getshortdoc, pad_console_length,
                   split_tags_from_doc, split_args_from_name_or_path)
from .typehints import copy_signature, KnownAtRuntime
from .unic import prepr, safe_str


def read_rest_data(rstfile):
    from .restreader import read_rest_data
    return read_rest_data(rstfile)


# Quietly deprecated utils. Should be deprecated loudly in RF 7.0.
# https://github.com/robotframework/robotframework/issues/4501

from .robottypes import FALSE_STRINGS, TRUE_STRINGS


# Deprecated Python 2/3 compatibility layer. Not needed by Robot Framework itself
# after RF 5.0 when Python 2 support was dropped. Should be deprecated loudly in
# RF 7.0. Notice that there's also `PY2` in the `platform` submodule.
# https://github.com/robotframework/robotframework/issues/4501

from io import StringIO


PY3 = True
PY2 = JYTHON = IRONPYTHON = False
is_unicode = is_string
unicode = str
unic = safe_str
roundup = round


def py2to3(cls):
    """Deprecated since RF 5.0. Use Python 3 features directly instead."""
    if hasattr(cls, '__unicode__'):
        cls.__str__ = lambda self: self.__unicode__()
    if hasattr(cls, '__nonzero__'):
        cls.__bool__ = lambda self: self.__nonzero__()
    return cls


def py3to2(cls):
    """Deprecated since RF 5.0. Never done anything when used on Python 3."""
    return cls



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
from contextlib import contextmanager
from pathlib import Path
from typing import cast, Iterator, Sequence, TextIO, Union

from robot.utils import file_writer, test_or_task

from .statements import (Break, Continue, ElseHeader, ElseIfHeader, End, ExceptHeader,
                         Error, FinallyHeader, ForHeader, IfHeader, KeywordCall,
                         KeywordName, Node, ReturnSetting, ReturnStatement,
                         SectionHeader, Statement, TemplateArguments, TestCaseName,
                         TryHeader, WhileHeader)
from .visitor import ModelVisitor
from ..lexer import Token


Body = Sequence[Union[Statement, 'Block']]
Errors = Sequence[str]


class Container(Node, ABC):

    @property
    def lineno(self) -> int:
        statement = FirstStatementFinder.find_from(self)
        return statement.lineno if statement else -1

    @property
    def col_offset(self) -> int:
        statement = FirstStatementFinder.find_from(self)
        return statement.col_offset if statement else -1

    @property
    def end_lineno(self) -> int:
        statement = LastStatementFinder.find_from(self)
        return statement.end_lineno if statement else -1

    @property
    def end_col_offset(self) -> int:
        statement = LastStatementFinder.find_from(self)
        return statement.end_col_offset if statement else -1

    def validate_model(self):
        ModelValidator().visit(self)

    def validate(self, ctx: 'ValidationContext'):
        pass


class File(Container):
    _fields = ('sections',)
    _attributes = ('source', 'languages') + Container._attributes

    def __init__(self, sections: 'Sequence[Section]' = (), source: 'Path|None' = None,
                 languages: Sequence[str] = ()):
        super().__init__()
        self.sections = list(sections)
        self.source = source
        self.languages = list(languages)

    def save(self, output: 'Path|str|TextIO|None' = None):
        """Save model to the given ``output`` or to the original source file.

        The ``output`` can be a path to a file or an already opened file
        object. If ``output`` is not given, the original source file will
        be overwritten.
        """
        output = output or self.source
        if output is None:
            raise TypeError('Saving model requires explicit output '
                            'when original source is not path.')
        ModelWriter(output).write(self)


class Block(Container, ABC):
    _fields = ('header', 'body')

    def __init__(self, header: 'Statement|None', body: Body = (), errors: Errors = ()):
        self.header = header
        self.body = list(body)
        self.errors = tuple(errors)

    def _body_is_empty(self):
        # This works with tests, keywords, and blocks inside them, not with sections.
        valid = (KeywordCall, TemplateArguments, Continue, Break, ReturnSetting,
                 ReturnStatement, NestedBlock, Error)
        return not any(isinstance(node, valid) for node in self.body)


class Section(Block):
    header: 'SectionHeader|None'


class SettingSection(Section):
    header: SectionHeader


class VariableSection(Section):
    header: SectionHeader


# TODO: should there be a separate TaskSection?
class TestCaseSection(Section):
    header: SectionHeader

    @property
    def tasks(self) -> bool:
        return self.header.type == Token.TASK_HEADER


class KeywordSection(Section):
    header: SectionHeader


class CommentSection(Section):
    header: 'SectionHeader|None'


class ImplicitCommentSection(CommentSection):
    header: None

    def __init__(self, header: 'Statement|None' = None, body: Body = (),
                 errors: Errors = ()):
        body = ([header] if header is not None else []) + list(body)
        super().__init__(None, body, errors)


class InvalidSection(Section):
    pass


class TestCase(Block):
    header: TestCaseName

    @property
    def name(self) -> str:
        return self.header.name

    def validate(self, ctx: 'ValidationContext'):
        if self._body_is_empty():
            self.errors += (test_or_task('{Test} cannot be empty.', ctx.tasks),)


class Keyword(Block):
    header: KeywordName

    @property
    def name(self) -> str:
        return self.header.name

    def validate(self, ctx: 'ValidationContext'):
        if self._body_is_empty():
            self.errors += ("User keyword cannot be empty.",)


class NestedBlock(Block):
    _fields = ('header', 'body', 'end')

    def __init__(self, header: Statement, body: Body = (), end: 'End|None' = None,
                 errors: Errors = ()):
        super().__init__(header, body, errors)
        self.end = end


class If(NestedBlock):
    """Represents IF structures in the model.

    Used with IF, Inline IF, ELSE IF and ELSE nodes. The :attr:`type` attribute
    specifies the type.
    """
    _fields = ('header', 'body', 'orelse', 'end')
    header: 'IfHeader|ElseIfHeader|ElseHeader'

    def __init__(self, header: Statement, body: Body = (), orelse: 'If|None' = None,
                 end: 'End|None' = None, errors: Errors = ()):
        super().__init__(header, body, end, errors)
        self.orelse = orelse

    @property
    def type(self) -> str:
        return self.header.type

    @property
    def condition(self) -> 'str|None':
        return self.header.condition

    @property
    def assign(self) -> 'tuple[str, ...]':
        return self.header.assign

    def validate(self, ctx: 'ValidationContext'):
        self._validate_body()
        if self.type == Token.IF:
            self._validate_structure()
            self._validate_end()
        if self.type == Token.INLINE_IF:
            self._validate_structure()
            self._validate_inline_if()

    def _validate_body(self):
        if self._body_is_empty():
            type = self.type if self.type != Token.INLINE_IF else 'IF'
            self.errors += (f'{type} branch cannot be empty.',)

    def _validate_structure(self):
        orelse = self.orelse
        else_seen = False
        while orelse:
            if else_seen:
                if orelse.type == Token.ELSE:
                    error = 'Only one ELSE allowed.'
                else:
                    error = 'ELSE IF not allowed after ELSE.'
                if error not in self.errors:
                    self.errors += (error,)
            else_seen = else_seen or orelse.type == Token.ELSE
            orelse = orelse.orelse

    def _validate_end(self):
        if not self.end:
            self.errors += ('IF must have closing END.',)

    def _validate_inline_if(self):
        branch = self
        assign = branch.assign
        while branch:
            if branch.body:
                item = cast(Statement, branch.body[0])
                if assign and item.type != Token.KEYWORD:
                    self.errors += ('Inline IF with assignment can only contain '
                                    'keyword calls.',)
                if getattr(item, 'assign', None):
                    self.errors += ('Inline IF branches cannot contain assignments.',)
                if item.type == Token.INLINE_IF:
                    self.errors += ('Inline IF cannot be nested.',)
            branch = branch.orelse


class For(NestedBlock):
    header: ForHeader

    @property
    def variables(self) -> 'tuple[str, ...]':
        return self.header.variables

    @property
    def values(self) -> 'tuple[str, ...]':
        return self.header.values

    @property
    def flavor(self) -> 'str|None':
        return self.header.flavor

    @property
    def start(self) -> 'str|None':
        return self.header.start

    @property
    def mode(self) -> 'str|None':
        return self.header.mode

    @property
    def fill(self) -> 'str|None':
        return self.header.fill

    def validate(self, ctx: 'ValidationContext'):
        if self._body_is_empty():
            self.errors += ('FOR loop cannot be empty.',)
        if not self.end:
            self.errors += ('FOR loop must have closing END.',)


class Try(NestedBlock):
    _fields = ('header', 'body', 'next', 'end')
    header: 'TryHeader|ExceptHeader|ElseHeader|FinallyHeader'

    def __init__(self, header: Statement, body: Body = (), next: 'Try|None' = None,
                 end: 'End|None' = None, errors: Errors = ()):
        super().__init__(header, body, end, errors)
        self.next = next

    @property
    def type(self) -> str:
        return self.header.type

    @property
    def patterns(self) -> 'tuple[str, ...]':
        return getattr(self.header, 'patterns', ())

    @property
    def pattern_type(self) -> 'str|None':
        return getattr(self.header, 'pattern_type', None)

    @property
    def variable(self) -> 'str|None':
        return getattr(self.header, 'variable', None)

    def validate(self, ctx: 'ValidationContext'):
        self._validate_body()
        if self.type == Token.TRY:
            self._validate_structure()
            self._validate_end()

    def _validate_body(self):
        if self._body_is_empty():
            self.errors += (f'{self.type} branch cannot be empty.',)

    def _validate_structure(self):
        else_count = 0
        finally_count = 0
        except_count = 0
        empty_except_count = 0
        branch = self.next
        while branch:
            if branch.type == Token.EXCEPT:
                if else_count:
                    self.errors += ('EXCEPT not allowed after ELSE.',)
                if finally_count:
                    self.errors += ('EXCEPT not allowed after FINALLY.',)
                if branch.patterns and empty_except_count:
                    self.errors += ('EXCEPT without patterns must be last.',)
                if not branch.patterns:
                    empty_except_count += 1
                except_count += 1
            if branch.type == Token.ELSE:
                if finally_count:
                    self.errors += ('ELSE not allowed after FINALLY.',)
                else_count += 1
            if branch.type == Token.FINALLY:
                finally_count += 1
            branch = branch.next
        if finally_count > 1:
            self.errors += ('Only one FINALLY allowed.',)
        if else_count > 1:
            self.errors += ('Only one ELSE allowed.',)
        if empty_except_count > 1:
            self.errors += ('Only one EXCEPT without patterns allowed.',)
        if not (except_count or finally_count):
            self.errors += ('TRY structure must have EXCEPT or FINALLY branch.',)

    def _validate_end(self):
        if not self.end:
            self.errors += ('TRY must have closing END.',)


class While(NestedBlock):
    header: WhileHeader

    @property
    def condition(self) -> str:
        return self.header.condition

    @property
    def limit(self) -> 'str|None':
        return self.header.limit

    @property
    def on_limit(self) -> 'str|None':
        return self.header.on_limit

    @property
    def on_limit_message(self) -> 'str|None':
        return self.header.on_limit_message

    def validate(self, ctx: 'ValidationContext'):
        if self._body_is_empty():
            self.errors += ('WHILE loop cannot be empty.',)
        if not self.end:
            self.errors += ('WHILE loop must have closing END.',)


class ModelWriter(ModelVisitor):

    def __init__(self, output: 'Path|str|TextIO'):
        if isinstance(output, (Path, str)):
            self.writer = file_writer(output)
            self.close_writer = True
        else:
            self.writer = output
            self.close_writer = False

    def write(self, model: Node):
        try:
            self.visit(model)
        finally:
            if self.close_writer:
                self.writer.close()

    def visit_Statement(self, statement: Statement):
        for token in statement:
            self.writer.write(token.value)


class ModelValidator(ModelVisitor):

    def __init__(self):
        self.ctx = ValidationContext()

    def visit_Block(self, node: Block):
        with self.ctx.block(node):
            node.validate(self.ctx)
            super().generic_visit(node)

    def visit_Statement(self, node: Statement):
        node.validate(self.ctx)


class ValidationContext:

    def __init__(self):
        self.blocks = []

    @contextmanager
    def block(self, node: Block) -> Iterator[None]:
        self.blocks.append(node)
        try:
            yield
        finally:
            self.blocks.pop()

    @property
    def parent_block(self) -> 'Block|None':
        return self.blocks[-1] if self.blocks else None

    @property
    def tasks(self) -> bool:
        for parent in self.blocks:
            if isinstance(parent, TestCaseSection):
                return parent.tasks
        return False

    @property
    def in_keyword(self) -> bool:
        return any(isinstance(b, Keyword) for b in self.blocks)

    @property
    def in_loop(self) -> bool:
        return any(isinstance(b, (For, While)) for b in self.blocks)

    @property
    def in_finally(self) -> bool:
        parent = self.parent_block
        return isinstance(parent, Try) and parent.header.type == Token.FINALLY


class FirstStatementFinder(ModelVisitor):

    def __init__(self):
        self.statement: 'Statement|None' = None

    @classmethod
    def find_from(cls, model: Node) -> 'Statement|None':
        finder = cls()
        finder.visit(model)
        return finder.statement

    def visit_Statement(self, statement: Statement):
        if self.statement is None:
            self.statement = statement

    def generic_visit(self, node: Node):
        if self.statement is None:
            super().generic_visit(node)


class LastStatementFinder(ModelVisitor):

    def __init__(self):
        self.statement: 'Statement|None' = None

    @classmethod
    def find_from(cls, model: Node) -> 'Statement|None':
        finder = cls()
        finder.visit(model)
        return finder.statement

    def visit_Statement(self, statement: Statement):
        self.statement = statement



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.utils import compress_text, html_format


class StringIndex(int):
    pass


class StringCache:
    _compress_threshold = 80
    _use_compressed_threshold = 1.1
    _zero_index = StringIndex(0)

    def __init__(self):
        self._cache = {('', False): self._zero_index}

    def add(self, text, html=False):
        if not text:
            return self._zero_index
        key = (text, html)
        if key not in self._cache:
            self._cache[key] = StringIndex(len(self._cache))
        return self._cache[key]

    def dump(self):
        return tuple(self._encode(text, html) for text, html in self._cache)

    def _encode(self, text, html=False):
        if html:
            text = html_format(text)
        if len(text) > self._compress_threshold:
            compressed = compress_text(text)
            if len(compressed) * self._use_compressed_threshold < len(text):
                return compressed
        # Strings starting with '*' are raw, others are compressed.
        return '*' + text



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import re

from .htmlformatters import LinkFormatter, HtmlFormatter


_format_url = LinkFormatter().format_url
_format_html = HtmlFormatter().format
_generic_escapes = (('&', '&amp;'), ('<', '&lt;'), ('>', '&gt;'))
_attribute_escapes = _generic_escapes \
         + (('"', '&quot;'), ('\n', '&#10;'), ('\r', '&#13;'), ('\t', '&#09;'))
_illegal_chars_in_xml = re.compile('[\x00-\x08\x0B\x0C\x0E-\x1F\uFFFE\uFFFF]')


def html_escape(text, linkify=True):
    text = _escape(text)
    if linkify and '://' in text:
        text = _format_url(text)
    return text


def xml_escape(text):
    return _illegal_chars_in_xml.sub('', _escape(text))


def html_format(text):
    return _format_html(_escape(text))


def attribute_escape(attr):
    attr = _escape(attr, _attribute_escapes)
    return _illegal_chars_in_xml.sub('', attr)


def _escape(text, escapes=_generic_escapes):
    for name, value in escapes:
        if name in text:  # performance optimization
            text = text.replace(name, value)
    return text



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys
from pprint import PrettyPrinter
from unicodedata import normalize


def safe_str(item):
    return normalize('NFC', _safe_str(item))


def _safe_str(item):
    if isinstance(item, str):
        return item
    if isinstance(item, (bytes, bytearray)):
        try:
            return item.decode('ASCII')
        except UnicodeError:
            return ''.join(chr(b) if b < 128 else '\\x%x' % b for b in item)
    try:
        return str(item)
    except:
        return _unrepresentable_object(item)


def prepr(item, width=80):
    return safe_str(PrettyRepr(width=width).pformat(item))


class PrettyRepr(PrettyPrinter):

    def format(self, object, context, maxlevels, level):
        try:
            return PrettyPrinter.format(self, object, context, maxlevels, level)
        except:
            return _unrepresentable_object(object), True, False

    # Don't split strings: https://stackoverflow.com/questions/31485402
    def _format(self, object, *args, **kwargs):
        if isinstance(object, (str, bytes, bytearray)):
            width = self._width
            self._width = sys.maxsize
            try:
                super()._format(object, *args, **kwargs)
            finally:
                self._width = width
        else:
            super()._format(object, *args, **kwargs)


def _unrepresentable_object(item):
    from .error import get_error_message
    return "<Unrepresentable object %s. Error: %s>" \
           % (item.__class__.__name__, get_error_message())



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import sys

from robot.errors import DataError
from robot.model import Message as BaseMessage
from robot.utils import get_timestamp, is_string, safe_str, console_encode


LEVELS = {
  'NONE'  : 7,
  'SKIP'  : 6,
  'FAIL'  : 5,
  'ERROR' : 4,
  'WARN'  : 3,
  'INFO'  : 2,
  'DEBUG' : 1,
  'TRACE' : 0,
}


def write_to_console(msg, newline=True, stream='stdout'):
    msg = str(msg)
    if newline:
        msg += '\n'
    stream = sys.__stdout__ if stream.lower() != 'stderr' else sys.__stderr__
    stream.write(console_encode(msg, stream=stream))
    stream.flush()


class AbstractLogger:

    def __init__(self, level='TRACE'):
        self._is_logged = IsLogged(level)

    def set_level(self, level):
        return self._is_logged.set_level(level)

    def trace(self, msg):
        self.write(msg, 'TRACE')

    def debug(self, msg):
        self.write(msg, 'DEBUG')

    def info(self, msg):
        self.write(msg, 'INFO')

    def warn(self, msg):
        self.write(msg, 'WARN')

    def fail(self, msg):
        html = False
        if msg.startswith("*HTML*"):
            html = True
            msg = msg[6:].lstrip()
        self.write(msg, 'FAIL', html)

    def skip(self, msg):
        html = False
        if msg.startswith("*HTML*"):
            html = True
            msg = msg[6:].lstrip()
        self.write(msg, 'SKIP', html)

    def error(self, msg):
        self.write(msg, 'ERROR')

    def write(self, message, level, html=False):
        self.message(Message(message, level, html))

    def message(self, msg):
        raise NotImplementedError(self.__class__)


class Message(BaseMessage):
    __slots__ = ['_message']

    def __init__(self, message, level='INFO', html=False, timestamp=None):
        message = self._normalize_message(message)
        level, html = self._get_level_and_html(level, html)
        timestamp = timestamp or get_timestamp()
        super().__init__(message, level, html, timestamp)

    def _normalize_message(self, msg):
        if callable(msg):
            return msg
        if not is_string(msg):
            msg = safe_str(msg)
        if '\r\n' in msg:
            msg = msg.replace('\r\n', '\n')
        return msg

    def _get_level_and_html(self, level, html):
        level = level.upper()
        if level == 'HTML':
            return 'INFO', True
        if level == 'CONSOLE':
            level = 'INFO'
        if level not in LEVELS:
            raise DataError("Invalid log level '%s'." % level)
        return level, html

    @property
    def message(self):
        self.resolve_delayed_message()
        return self._message

    @message.setter
    def message(self, message):
        self._message = message

    def resolve_delayed_message(self):
        if callable(self._message):
            self._message = self._message()


class IsLogged:

    def __init__(self, level):
        self.level = level.upper()
        self._int_level = self._level_to_int(level)

    def __call__(self, level):
        return self._level_to_int(level) >= self._int_level

    def set_level(self, level):
        old = self.level
        self.__init__(level)
        return old

    def _level_to_int(self, level):
        try:
            return LEVELS[level.upper()]
        except KeyError:
            raise DataError("Invalid log level '%s'." % level)


class AbstractLoggerProxy:
    _methods = None
    _no_method = lambda *args: None

    def __init__(self, logger, method_names=None, prefix=None):
        self.logger = logger
        for name in method_names or self._methods:
            # Allow extending classes to implement some of the messages themselves.
            if hasattr(self, name):
                if hasattr(logger, name):
                    continue
                target = logger
            else:
                target = self
            setattr(target, name, self._get_method(logger, name, prefix))

    def _get_method(self, logger, name, prefix):
        for method_name in self._get_method_names(name, prefix):
            if hasattr(logger, method_name):
                return getattr(logger, method_name)
        return self._no_method

    def _get_method_names(self, name, prefix):
        names = [name, self._toCamelCase(name)] if '_' in name else [name]
        if prefix:
            names += [prefix + name for name in names]
        return names

    def _toCamelCase(self, name):
        parts = name.split('_')
        return ''.join([parts[0]] + [part.capitalize() for part in parts[1:]])



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.output.loggerhelper import IsLogged

from robot.model import SuiteVisitor


class MessageFilter(SuiteVisitor):

    def __init__(self, log_level=None):
        self.is_logged = IsLogged(log_level or 'TRACE')

    def start_suite(self, suite):
        if self.is_logged.level == 'TRACE':
            return False

    def start_keyword(self, keyword):
        for item in list(keyword.body):
            if item.type == item.MESSAGE and not self.is_logged(item.level):
                keyword.body.remove(item)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.result import ResultVisitor
from robot.utils import XmlWriter


class XUnitWriter:

    def __init__(self, execution_result):
        self._execution_result = execution_result

    def write(self, output):
        xml_writer = XmlWriter(output, usage='xunit')
        writer = XUnitFileWriter(xml_writer)
        self._execution_result.visit(writer)


class XUnitFileWriter(ResultVisitor):
    """Provides an xUnit-compatible result file.

    Attempts to adhere to the de facto schema guessed by Peter Reilly, see:
    http://marc.info/?l=ant-dev&m=123551933508682
    """

    def __init__(self, xml_writer):
        self._writer = xml_writer

    def start_suite(self, suite):
        stats = suite.statistics  # Accessing property only once.
        attrs = {'name': suite.name,
                 'tests': f'{stats.total}',
                 'errors': '0',
                 'failures': f'{stats.failed}',
                 'skipped': f'{stats.skipped}',
                 'time': self._time_as_seconds(suite.elapsedtime),
                 'timestamp' : self._starttime_to_isoformat(suite.starttime)}
        self._writer.start('testsuite', attrs)

    def end_suite(self, suite):
        if suite.metadata or suite.doc:
            self._writer.start('properties')
            if suite.doc:
                self._writer.element('property', attrs={'name': 'Documentation', 'value': suite.doc})
            for meta_name, meta_value in suite.metadata.items():
                self._writer.element('property', attrs={'name': meta_name, 'value': meta_value})
            self._writer.end('properties')
        self._writer.end('testsuite')

    def visit_test(self, test):
        self._writer.start('testcase',
                           {'classname': test.parent.longname,
                            'name': test.name,
                            'time': self._time_as_seconds(test.elapsedtime)})
        if test.failed:
            self._writer.element('failure', attrs={'message': test.message,
                                                   'type': 'AssertionError'})
        if test.skipped:
            self._writer.element('skipped', attrs={'message': test.message,
                                                   'type': 'SkipExecution'})
        self._writer.end('testcase')

    def _time_as_seconds(self, millis):
        return format(millis / 1000, '.3f')

    def visit_keyword(self, kw):
        pass

    def visit_statistics(self, stats):
        pass

    def visit_errors(self, errors):
        pass

    def end_result(self, result):
        self._writer.close()

    def _starttime_to_isoformat(self, stime):
        if not stime:
            return None
        return f'{stime[:4]}-{stime[4:6]}-{stime[6:8]}T{stime[9:22]}000'



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Optional base classes for libraries and other extensions.

Module contents:

- :class:`DynamicLibrary` for libraries using the `dynamic library API`__.
- :class:`HybridLibrary` for libraries using the `hybrid library API`__.
- :class:`ListenerV2` for `listener interface version 2`__.
- :class:`ListenerV3` for `listener interface version 3`__.
- :class:`Parser` for `custom parsers`__. Also
  :class:`~robot.running.builder.settings.TestDefaults` used in ``Parser``
  type hints can be imported via this module if needed.
- Type definitions used by the aforementioned classes.

Main benefit of using these base classes is that editors can provide automatic
completion, documentation and type information. Their usage is not required.
Notice also that libraries typically use the static API and do not need any
base class.

.. note:: These classes are not exposed via the top level :mod:`robot.api`
          package and need to imported via :mod:`robot.api.interfaces`.

.. note:: Using this module requires having the typing_extensions__ module
          installed when using Python 3.6 or 3.7.

This module is new in Robot Framework 6.1.

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#dynamic-library-api
__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#hybrid-library-api
__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#listener-version-2
__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#listener-version-3
__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#parser-interface
__ https://pypi.org/project/typing-extensions/
"""

import sys
from abc import ABC, abstractmethod
from pathlib import Path
from typing import Any, Dict, List, Optional, Sequence, Tuple, Union
# Need to use version check and not try/except to support Mypy's stubgen.
if sys.version_info >= (3, 8):
    from typing import TypedDict
else:
    try:
        from typing_extensions import TypedDict
    except ImportError:
        raise ImportError("Using the 'robot.api.interfaces' module requires having "
                          "the 'typing_extensions' module installed with Python < 3.8.")
if sys.version_info >= (3, 10):
    from types import UnionType
else:
    UnionType = type

from robot import result, running
from robot.model import Message
from robot.running import TestDefaults, TestSuite


# Type aliases used by DynamicLibrary and HybridLibrary.
Name = str
PositArgs = List[Any]
NamedArgs = Dict[str, Any]
Documentation = str
Arguments = List[
    Union[
        str,               # Name with possible default like `arg` or `arg=1`.
        Tuple[str],        # Name without a default like `('arg',)`.
        Tuple[str, Any]    # Name and default like `('arg', 1)`.
    ]
]
Type = Union[
    type,                           # Actual type.
    str,                            # Type name or alias.
    UnionType,                      # Union syntax (e.g. `int | float`).
    Tuple[Union[type, str], ...]    # Tuple of types. Behaves like union.
]
Types = Union[
    Dict[str, Type],       # Types by name.
    List[                  # Types by position.
        Union[
            Type,          # Type info.
            None           # No type info.
        ]
    ]
]
Tags = List[str]
Source = str


class DynamicLibrary(ABC):
    """Optional base class for libraries using the dynamic library API.

    The dynamic library API makes it possible to dynamically specify
    what keywords a library implements and run them by using
    :meth:`get_keyword_names` and :meth:`run_keyword` methods, respectively.
    In addition to that it has various optional methods for returning more
    information about the implemented keywords to Robot Framework.
    """

    @abstractmethod
    def get_keyword_names(self) -> List[Name]:
        """Return names of the keywords this library implements.

        :return: Keyword names as a list of strings.

        ``name`` passed to other methods is always in the same format as
        returned by this method.
        """
        raise NotImplementedError

    @abstractmethod
    def run_keyword(self, name: Name, args: PositArgs, named: NamedArgs) -> Any:
        """Execute the specified keyword using the given arguments.

        :param name: Keyword name as a string.
        :param args: Positional arguments as a list.
        :param named: Named arguments as a dictionary.
        :raises: Reporting FAIL or SKIP status.
        :return: Keyword's return value.

        Reporting status, logging, returning values, etc. is handled the same
        way as with the normal static library API.
        """
        raise NotImplementedError

    def get_keyword_documentation(self, name: Name) -> Optional[Documentation]:
        """Optional method to return keyword documentation.

        The first logical line of keyword documentation is shown in
        the execution log under the executed keyword. The whole
        documentation is shown in documentation generated by Libdoc.

        :param name: Keyword name as a string.
        :return: Documentation as a string oras ``None`` if there is no
            documentation.

        This method is also used to get the overall library documentation as
        well as documentation related to importing the library. They are
        got by calling this method with special names ``__intro__`` and
        ``__init__``, respectively.
        """
        return None

    def get_keyword_arguments(self, name: Name) -> Optional[Arguments]:
        """Optional method to return keyword's argument specification.

        Returned information is used during execution for argument validation.
        In addition to that, arguments are shown in documentation generated
        by Libdoc.

        :param name: Keyword name as a string.
        :return: Argument specification using format explained below.

        Argument specification defines what arguments the keyword accepts.
        Returning ``None`` means that the keywords accepts any arguments.
        Accepted arguments are returned as a list using these rules:

        - Normal arguments are specified as a list of strings like
          ``['arg1', 'arg2']``. An empty list denotes that the keyword
          accepts no arguments.
        - Varargs must have a ``*`` prefix like ``['*numbers']``. There can
          be only one varargs, and it must follow normal arguments.
        - Arguments after varargs like ``['*items', 'arg']`` are considered
          named-only arguments.
        - If keyword does not accept varargs, a lone ``*`` can be used
          a separator between normal and named-only arguments like
          ``['normal', '*', 'named']``.
        - Kwargs must have a ``**``  prefix like ``['**config']``. There can
          be only one kwargs, and it must be last.

        Both normal arguments and named-only arguments can have default values:

        - Default values can be embedded to argument names so that they are
          separated with the equal sign like ``name=default``. In this case
          the default value type is always a string.
        - Alternatively arguments and their default values can be represented
          as two-tuples like ``('name', 'default')``. This allows non-string
          default values and automatic argument conversion based on them.
        - Arguments without default values can also be specified as tuples
          containing just the name like ``('name',)``.
        - With normal arguments, arguments with default values must follow
          arguments without them. There is no such restriction with named-only
          arguments.
        """
        return None

    def get_keyword_types(self, name: Name) -> Optional[Types]:
        """Optional method to return keyword's type specification.

        Type information is used for automatic argument conversion during
        execution. It is also shown in documentation generated by Libdoc.

        :param name: Keyword name as a string.
        :return: Type specification as a dictionary, as a list, or as ``None``
            if type information is not known.

        Type information can be mapped to arguments returned by
        :meth:`get_keyword_names` either by names using a dictionary or
        by position using a list. For example, if a keyword has argument
        specification ``['arg', 'second']``, it would be possible to return
        types both like ``{'arg': str, 'second': int}`` and ``[str, int]``.

        Regardless of the approach that is used, it is not necessarily to
        specify types for all arguments. When using a dictionary, some
        arguments can be omitted altogether. When using a list, it is possible
        to use ``None`` to mark that a certain argument does not have type
        information and arguments at the end can be omitted altogether.

        If is possible to specify that an argument has multiple possible types
        by using unions like ``{'arg': Union[int, float]}`` or tuples like
        ``{'arg': (int, float)}``.

        In addition to specifying types using classes, it is also possible
        to use names or aliases like ``{'a': 'int', 'b': 'boolean'}``.
        For an up-to-date list of supported types, names and aliases see
        the User Guide.
        """
        return None

    def get_keyword_tags(self, name: Name) -> Optional[Tags]:
        """Optional method to return keyword's tags.

        Tags are shown in the execution log and in documentation generated by
        Libdoc. Tags can also be used with various command line options.

        :param name: Keyword name as a string.
        :return: Tags as a list of strings or ``None`` if there are no tags.
        """
        return None

    def get_keyword_source(self, name: Name) -> Optional[Source]:
        """Optional method to return keyword's source path and line number.

        Source information is used by IDEs to provide navigation from
        keyword usage to implementation.

        :param name: Keyword name as a string.
        :return: Source as a string in format ``path:lineno`` or ``None``
            if source is not known.

        The general format to return the source is ``path:lineno`` like
        ``/example/Lib.py:42``. If the line number is not known, it is
        possible to return only the path. If the keyword is in the same
        file as the main library class, the path can be omitted and only
        the line number returned like ``:42``.

        The source information of the library itself is got automatically from
        the imported library class. The library source path is used with all
        keywords that do not return their own path.
        """
        return None


class HybridLibrary(ABC):
    """Optional base class for libraries using the hybrid library API.

    Hybrid library API makes it easy to specify what keywords a library
    implements by using the :meth:`get_keyword_names` method. After getting
    keyword names, Robot Framework uses ``getattr`` to get the actual keyword
    methods exactly like it does when using the normal static library API.
    Keyword name, arguments, documentation, tags, and so on are got directly
    from the keyword method.

    It is possible to implement keywords also outside the main library class.
    In such cases the library needs to have a ``__getattr__`` method that
    returns desired keyword methods.
    """

    @abstractmethod
    def get_keyword_names(self) -> List[Name]:
        """Return names of the implemented keyword methods as a list or strings.

        Returned names must match names of the implemented keyword methods.
        """
        raise NotImplementedError


# Attribute dictionary specifications used by ListenerV2.

class StartSuiteAttributes(TypedDict):
    """Attributes passed to listener v2 ``start_suite`` method.

    See the User Guide for more information.
    """
    id: str
    longname: str
    doc: str
    metadata: dict
    source: str
    suites: List[str]
    tests: List[str]
    totaltests: int
    starttime: str


class EndSuiteAttributes(StartSuiteAttributes):
    """Attributes passed to listener v2 ``end_suite`` method.

    See the User Guide for more information.
    """
    endtime: str
    elapsedtime: int
    status: str
    statistics: str
    message: str


class StartTestAttributes(TypedDict):
    """Attributes passed to listener v2 ``start_test`` method.

    See the User Guide for more information.
    """
    id: str
    longname: str
    originalname: str
    doc: str
    tags: List[str]
    template: str
    source: str
    lineno: int
    starttime: str


class EndTestAttributes(StartTestAttributes):
    """Attributes passed to listener v2 ``end_test`` method.

    See the User Guide for more information.
    """
    endtime: str
    elapedtime: int
    status: str
    message: str


class OptionalKeywordAttributes(TypedDict, total=False):
    """Extra attributes passed to listener v2 ``start/end_keyword`` methods.

    These attributes are included with control structures. For example, with
    IF structures attributes include ``condition``.
    """
    # FOR
    variables: List[str]
    flavor: str
    values: List[str]
    # ITERATION with FOR
    variables: Dict[str, str]
    # WHILE and IF
    condition: str
    # WHILE
    limit: str
    # EXCEPT
    patterns: List[str]
    pattern_type: str
    variable: str
    # RETURN
    values: List[str]


class StartKeywordAttributes(OptionalKeywordAttributes):
    """Attributes passed to listener v2 ``start_keyword`` method.

    See the User Guide for more information.
    """
    type: str
    kwname: str
    libname: str
    doc: str
    args: List[str]
    assign: List[str]
    tags: List[str]
    source: str
    lineno: int
    status: str
    starttime: str


class EndKeywordAttributes(StartKeywordAttributes):
    """Attributes passed to listener v2 ``end_keyword`` method.

    See the User Guide for more information.
    """
    endtime: str
    elapsedtime: int


class MessageAttributes(TypedDict):
    """Attributes passed to listener v2 ``log_message`` and ``messages`` methods.

    See the User Guide for more information.
    """
    message: str
    level: str
    timestamp: str
    html: str


class LibraryAttributes(TypedDict):
    """Attributes passed to listener v2 ``library_import`` method.

    See the User Guide for more information.
    """
    args: List[str]
    originalname: str
    source: str
    importer: Union[str, None]


class ResourceAttributes(TypedDict):
    """Attributes passed to listener v2 ``resource_import`` method.

    See the User Guide for more information.
    """
    source: str
    importer: Union[str, None]


class VariablesAttributes(TypedDict):
    """Attributes passed to listener v2 ``variables_import`` method.

    See the User Guide for more information.
    """
    args: List[str]
    source: str
    importer: Union[str, None]


class ListenerV2:
    """Optional base class for listeners using the listener API v2."""
    ROBOT_LISTENER_API_VERSION = 2

    def start_suite(self, name: str, attributes: StartSuiteAttributes):
        """Called when a suite starts."""

    def end_suite(self, name: str, attributes: EndSuiteAttributes):
        """Called when a suite end."""

    def start_test(self, name: str, attributes: StartTestAttributes):
        """Called when a test or task starts."""

    def end_test(self, name: str, attributes: EndTestAttributes):
        """Called when a test or task ends."""

    def start_keyword(self, name: str, attributes: StartKeywordAttributes):
        """Called when a keyword or a control structure like IF starts.

        The type of the started item is in ``attributes['type']``. Control
        structures can contain extra attributes that are only relevant to them.
        """

    def end_keyword(self, name: str, attributes: EndKeywordAttributes):
        """Called when a keyword or a control structure like IF ends.

        The type of the started item is in ``attributes['type']``. Control
        structures can contain extra attributes that are only relevant to them.
        """

    def log_message(self, message: MessageAttributes):
        """Called when a normal log message are emitted.

        The messages are typically logged by keywords, but also the framework
        itself logs some messages. These messages end up to output.xml and
        log.html.
        """

    def message(self, message: MessageAttributes):
        """Called when framework's internal messages are emitted.

        Only logged by the framework itself. These messages end up to the syslog
        if it is enabled.
        """

    def library_import(self, name: str, attributes: LibraryAttributes):
        """Called after a library has been imported."""

    def resource_import(self, name: str, attributes: ResourceAttributes):
        """Called after a resource file has been imported."""

    def variables_import(self, name: str, attributes: VariablesAttributes):
        """Called after a variable file has been imported."""

    def output_file(self, path: str):
        """Called after the output file has been created.

        At this point the file is guaranteed to be closed.
        """

    def log_file(self, path: str):
        """Called after the log file has been created."""

    def report_file(self, path: str):
        """Called after the report file has been created."""

    def xunit_file(self, path: str):
        """Called after the xunit compatible output file has been created."""

    def debug_file(self, path: str):
        """Called after the debug file has been created."""

    def close(self):
        """Called when the whole execution ends.

        With library listeners called when the library goes out of scope.
        """


class ListenerV3:
    """Optional base class for listeners using the listener API v3."""
    ROBOT_LISTENER_API_VERSION = 3

    def start_suite(self, data: running.TestSuite, result: result.TestSuite):
        """Called when a suite starts."""

    def end_suite(self, data: running.TestSuite, result: result.TestSuite):
        """Called when a suite ends."""

    def start_test(self, data: running.TestCase, result: result.TestCase):
        """Called when a test or task starts."""

    def end_test(self, data: running.TestCase, result: result.TestCase):
        """Called when a test or ends starts."""

    def log_message(self, message: Message):
        """Called when a normal log message are emitted.

        The messages are typically logged by keywords, but also the framework
        itself logs some messages. These messages end up to output.xml and
        log.html.
        """

    def message(self, message: Message):
        """Called when framework's internal messages are emitted.

        Only logged by the framework itself. These messages end up to the syslog
        if it is enabled.
        """

    def output_file(self, path: str):
        """Called after the output file has been created.

        At this point the file is guaranteed to be closed.
        """

    def log_file(self, path: str):
        """Called after the log file has been created."""

    def report_file(self, path: str):
        """Called after the report file has been created."""

    def xunit_file(self, path: str):
        """Called after the xunit compatible output file has been created."""

    def debug_file(self, path: str):
        """Called after the debug file has been created."""

    def close(self):
        """Called when the whole execution ends.

        With library listeners called when the library goes out of scope.
        """


class Parser(ABC):
    """Optional base class for custom parsers.

    Parsers do not need to explicitly extend this class and in simple cases
    it is possible to implement them as modules. Regardless how a parser is
    implemented, it must have :attr:`extension` attribute and :meth:`parse`
    method. The :meth:`parse_init` method is optional and only needed if
    a parser supports parsing suite initialization files.

    The mandatory :attr:`extension` attribute specifies what file extension or
    extensions a parser supports. It can be set either as a class or instance
    attribute, and it can be either a string or a sequence of strings. The
    attribute can also be named ``EXTENSION``, which typically works better
    when a parser is implemented as a module.

    Example::

        from pathlib import Path
        from robot.api import TestSuite
        from robot.api.interfaces import Parser, TestDefaults


        class ExampleParser(Parser):
            extension = '.example'

            def parse(self, source: Path, defaults: TestDefaults) -> TestSuite:
                suite = TestSuite(TestSuite.name_from_source(source), source=source)
                # parse the source file and add tests to the created suite
                return suite

    The support for custom parsers is new in Robot Framework 6.1.
    """
    extension: Union[str, Sequence[str]]

    @abstractmethod
    def parse(self, source: Path, defaults: TestDefaults) -> TestSuite:
        """Mandatory method for parsing suite files.

        :param source: Path to the file to parse.
        :param defaults: Default values set for test in init files.

        The ``defaults`` argument is optional. It is possible to implement
        this method also so that it accepts only ``source``.
        """
        raise NotImplementedError

    def parse_init(self, source: Path, defaults: TestDefaults) -> TestSuite:
        """Optional method for parsing suite initialization files.

        :param source: Path to the file to parse.
        :param defaults: Default values to used with tests in child suites.

        The ``defaults`` argument is optional. It is possible to implement
        this method also so that it accepts only ``source``.

        If this method is not implemented, possible initialization files cause
        an error.
        """
        raise NotImplementedError



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from functools import partial
import inspect
import os

from robot.errors import DataError
from robot.libraries import STDLIBS
from robot.output import LOGGER
from robot.utils import (getdoc, get_error_details, Importer, is_dict_like, is_init,
                         is_list_like, normalize, seq2str2, type_name)

from .arguments import EmbeddedArguments, CustomArgumentConverters
from .context import EXECUTION_CONTEXTS
from .dynamicmethods import (GetKeywordArguments, GetKeywordDocumentation,
                             GetKeywordNames, GetKeywordTags, RunKeyword)
from .handlers import Handler, InitHandler, DynamicHandler, EmbeddedArgumentsHandler
from .handlerstore import HandlerStore
from .libraryscopes import LibraryScope
from .outputcapture import OutputCapturer


def TestLibrary(name, args=None, variables=None, create_handlers=True, logger=LOGGER):
    if name in STDLIBS:
        import_name = 'robot.libraries.' + name
    else:
        import_name = name
    with OutputCapturer(library_import=True):
        importer = Importer('library', logger=LOGGER)
        libcode, source = importer.import_class_or_module(import_name,
                                                          return_source=True)
    libclass = _get_lib_class(libcode)
    lib = libclass(libcode, name, args or [], source, logger, variables)
    if create_handlers:
        lib.create_handlers()
    return lib


def _get_lib_class(libcode):
    if inspect.ismodule(libcode):
        return _ModuleLibrary
    if GetKeywordNames(libcode):
        if RunKeyword(libcode):
            return _DynamicLibrary
        else:
            return _HybridLibrary
    return _ClassLibrary


class _BaseTestLibrary:
    get_handler_error_level = 'INFO'

    def __init__(self, libcode, name, args, source, logger, variables):
        if os.path.exists(name):
            name = os.path.splitext(os.path.basename(os.path.abspath(name)))[0]
        self._libcode = libcode
        self._libinst = None
        self.version = self._get_version(libcode)
        self.name = name
        self.orig_name = name  # Stores original name when importing WITH NAME
        self.source = source
        self.logger = logger
        self.converters = self._get_converters(libcode)
        self.handlers = HandlerStore()
        self.has_listener = None  # Set when first instance is created
        self._doc = None
        self.doc_format = self._get_doc_format(libcode)
        self.scope = LibraryScope(libcode, self)
        self.init = self._create_init_handler(libcode)
        self.positional_args, self.named_args \
            = self.init.resolve_arguments(args, variables)

    def __len__(self):
        return len(self.handlers)

    def __bool__(self):
        return bool(self.handlers) or self.has_listener

    @property
    def doc(self):
        if self._doc is None:
            self._doc = getdoc(self.get_instance())
        return self._doc

    @property
    def lineno(self):
        if inspect.ismodule(self._libcode):
            return 1
        try:
            lines, start_lineno = inspect.getsourcelines(self._libcode)
        except (TypeError, OSError, IOError):
            return -1
        for increment, line in enumerate(lines):
            if line.strip().startswith('class '):
                return start_lineno + increment
        return start_lineno

    def create_handlers(self):
        self._create_handlers(self.get_instance())
        self.reset_instance()

    def handlers_for(self, name):
        return self.handlers.get_handlers(name)

    def reload(self):
        self.handlers = HandlerStore()
        self._create_handlers(self.get_instance())

    def start_suite(self):
        self.scope.start_suite()

    def end_suite(self):
        self.scope.end_suite()

    def start_test(self):
        self.scope.start_test()

    def end_test(self):
        self.scope.end_test()

    def report_error(self, message, details=None, level='ERROR',
                     details_level='INFO'):
        prefix = 'Error in' if level in ('ERROR', 'WARN') else 'In'
        self.logger.write("%s library '%s': %s" % (prefix, self.name, message),
                          level)
        if details:
            self.logger.write('Details:\n%s' % details, details_level)

    def _get_version(self, libcode):
        return self._get_attr(libcode, 'ROBOT_LIBRARY_VERSION') \
            or self._get_attr(libcode, '__version__')

    def _get_attr(self, object, attr, default='', upper=False):
        value = str(getattr(object, attr, default))
        if upper:
            value = normalize(value, ignore='_').upper()
        return value

    def _get_doc_format(self, libcode):
        return self._get_attr(libcode, 'ROBOT_LIBRARY_DOC_FORMAT', upper=True)

    def _create_init_handler(self, libcode):
        return InitHandler(self, self._resolve_init_method(libcode))

    def _resolve_init_method(self, libcode):
        init = getattr(libcode, '__init__', None)
        return init if is_init(init) else None

    def _get_converters(self, libcode):
        converters = getattr(libcode, 'ROBOT_LIBRARY_CONVERTERS', None)
        if not converters:
            return None
        if not is_dict_like(converters):
            self.report_error(f'Argument converters must be given as a dictionary, '
                              f'got {type_name(converters)}.')
            return None
        return CustomArgumentConverters.from_dict(converters, self)

    def reset_instance(self, instance=None):
        prev = self._libinst
        if not self.scope.is_global:
            self._libinst = instance
        return prev

    def get_instance(self, create=True):
        if not create:
            return self._libinst
        if self._libinst is None:
            self._libinst = self._get_instance(self._libcode)
        if self.has_listener is None:
            self.has_listener = bool(self.get_listeners(self._libinst))
        return self._libinst

    def _get_instance(self, libcode):
        with OutputCapturer(library_import=True):
            try:
                return libcode(*self.positional_args, **dict(self.named_args))
            except:
                self._raise_creating_instance_failed()

    def get_listeners(self, libinst=None):
        if libinst is None:
            libinst = self.get_instance()
        listeners = getattr(libinst, 'ROBOT_LIBRARY_LISTENER', None)
        if listeners is None:
            return []
        if is_list_like(listeners):
            return listeners
        return [listeners]

    def register_listeners(self):
        if self.has_listener:
            try:
                listeners = EXECUTION_CONTEXTS.current.output.library_listeners
                listeners.register(self.get_listeners(), self)
            except DataError as err:
                self.has_listener = False
                # Error should have information about suite where the
                # problem occurred but we don't have such info here.
                self.report_error("Registering listeners failed: %s" % err)

    def unregister_listeners(self, close=False):
        if self.has_listener:
            listeners = EXECUTION_CONTEXTS.current.output.library_listeners
            listeners.unregister(self, close)

    def close_global_listeners(self):
        if self.scope.is_global:
            for listener in self.get_listeners():
                self._close_listener(listener)

    def _close_listener(self, listener):
        method = (getattr(listener, 'close', None) or
                  getattr(listener, '_close', None))
        try:
            if method:
                method()
        except Exception:
            message, details = get_error_details()
            name = getattr(listener, '__name__', None) or type_name(listener)
            self.report_error("Calling method '%s' of listener '%s' failed: %s"
                              % (method.__name__, name, message), details)

    def _create_handlers(self, libcode):
        try:
            names = self._get_handler_names(libcode)
        except Exception:
            message, details = get_error_details()
            raise DataError("Getting keyword names from library '%s' failed: %s"
                            % (self.name, message), details)
        for name in names:
            method = self._try_to_get_handler_method(libcode, name)
            if method:
                handler, embedded = self._try_to_create_handler(name, method)
                if handler:
                    try:
                        self.handlers.add(handler, embedded)
                    except DataError as err:
                        self._adding_keyword_failed(handler.name, err)
                    else:
                        self.logger.debug("Created keyword '%s'" % handler.name)

    def _get_handler_names(self, libcode):
        def has_robot_name(name):
            try:
                handler = self._get_handler_method(libcode, name)
            except DataError:
                return False
            return hasattr(handler, 'robot_name')

        auto_keywords = getattr(libcode, 'ROBOT_AUTO_KEYWORDS', True)
        if auto_keywords:
            predicate = lambda name: name[:1] != '_' or has_robot_name(name)
        else:
            predicate = has_robot_name
        return [name for name in dir(libcode) if predicate(name)]

    def _try_to_get_handler_method(self, libcode, name):
        try:
            return self._get_handler_method(libcode, name)
        except DataError as err:
            self._adding_keyword_failed(name, err, self.get_handler_error_level)
            return None

    def _adding_keyword_failed(self, name, error, level='ERROR'):
        self.report_error(
            "Adding keyword '%s' failed: %s" % (name, error.message),
            error.details,
            level=level,
            details_level='DEBUG'
        )

    def _get_handler_method(self, libcode, name):
        try:
            method = getattr(libcode, name)
        except Exception:
            message, details = get_error_details()
            raise DataError(f'Getting handler method failed: {message}', details)
        return self._validate_handler_method(method)

    def _validate_handler_method(self, method):
        # isroutine returns false for partial objects. This may change in the future.
        if not (inspect.isroutine(method) or isinstance(method, partial)):
            raise DataError('Not a method or function.')
        if getattr(method, 'robot_not_keyword', False):
            raise DataError('Not exposed as a keyword.')
        return method

    def _try_to_create_handler(self, name, method):
        try:
            handler = self._create_handler(name, method)
        except DataError as err:
            self._adding_keyword_failed(name, err)
            return None, False
        try:
            return self._get_possible_embedded_args_handler(handler)
        except DataError as err:
            self._adding_keyword_failed(handler.name, err)
            return None, False

    def _create_handler(self, handler_name, handler_method):
        return Handler(self, handler_name, handler_method)

    def _get_possible_embedded_args_handler(self, handler):
        embedded = EmbeddedArguments.from_name(handler.name)
        if embedded:
            self._validate_embedded_count(embedded, handler.arguments)
            return EmbeddedArgumentsHandler(embedded, handler), True
        return handler, False

    def _validate_embedded_count(self, embedded, arguments):
        if not (arguments.minargs <= len(embedded.args) <= arguments.maxargs):
            raise DataError('Embedded argument count does not match number of '
                            'accepted arguments.')

    def _raise_creating_instance_failed(self):
        msg, details = get_error_details()
        if self.positional_args or self.named_args:
            args = self.positional_args + ['%s=%s' % item for item in self.named_args]
            args_text = 'arguments %s' % seq2str2(args)
        else:
            args_text = 'no arguments'
        raise DataError("Initializing library '%s' with %s failed: %s\n%s"
                        % (self.name, args_text, msg, details))


class _ClassLibrary(_BaseTestLibrary):

    def _get_handler_method(self, libinst, name):
        for item in (libinst,) + inspect.getmro(libinst.__class__):
            # `isroutine` is used before `getattr` to avoid calling properties.
            if (name in getattr(item, '__dict__', ())
                    and inspect.isroutine(item.__dict__[name])):
                try:
                    method = getattr(libinst, name)
                except Exception:
                    message, traceback = get_error_details()
                    raise DataError(f'Getting handler method failed: {message}',
                                    traceback)
                return self._validate_handler_method(method)
        raise DataError('Not a method or function.')


class _ModuleLibrary(_BaseTestLibrary):

    def _get_handler_method(self, libcode, name):
        method = _BaseTestLibrary._get_handler_method(self, libcode, name)
        if hasattr(libcode, '__all__') and name not in libcode.__all__:
            raise DataError('Not exposed as a keyword.')
        return method

    def get_instance(self, create=True):
        if not create:
            return self._libcode
        if self.has_listener is None:
            self.has_listener = bool(self.get_listeners(self._libcode))
        return self._libcode

    def _create_init_handler(self, libcode):
        return InitHandler(self)


class _HybridLibrary(_BaseTestLibrary):
    get_handler_error_level = 'ERROR'

    def _get_handler_names(self, instance):
        return GetKeywordNames(instance)()


class _DynamicLibrary(_BaseTestLibrary):
    get_handler_error_level = 'ERROR'

    def __init__(self, libcode, name, args, source, logger, variables=None):
        _BaseTestLibrary.__init__(self, libcode, name, args, source, logger,
                                  variables)

    @property
    def doc(self):
        if self._doc is None:
            self._doc = (self._get_kw_doc('__intro__') or
                         _BaseTestLibrary.doc.fget(self))
        return self._doc

    def _get_kw_doc(self, name):
        getter = GetKeywordDocumentation(self.get_instance())
        return getter(name)

    def _get_kw_args(self, name):
        getter = GetKeywordArguments(self.get_instance())
        return getter(name)

    def _get_kw_tags(self, name):
        getter = GetKeywordTags(self.get_instance())
        return getter(name)

    def _get_handler_names(self, instance):
        return GetKeywordNames(instance)()

    def _get_handler_method(self, instance, name):
        return RunKeyword(instance)

    def _create_handler(self, name, method):
        argspec = self._get_kw_args(name)
        tags = self._get_kw_tags(name)
        doc = self._get_kw_doc(name)
        return DynamicHandler(self, name, method, doc, argspec, tags)

    def _create_init_handler(self, libcode):
        docgetter = lambda: self._get_kw_doc('__init__')
        return InitHandler(self, self._resolve_init_method(libcode), docgetter)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.errors import DataError
from robot.utils import (is_dict_like, is_list_like, plural_or_not as s,
                         seq2str, type_name)


class TypeValidator:

    def __init__(self, argspec):
        """:type argspec: :py:class:`robot.running.arguments.ArgumentSpec`"""
        self._argspec = argspec

    def validate(self, types):
        if types is None:
            return None
        if not types:
            return {}
        if is_dict_like(types):
            return self.validate_type_dict(types)
        if is_list_like(types):
            return self.convert_type_list_to_dict(types)
        raise DataError('Type information must be given as a dictionary or '
                        'a list, got %s.' % type_name(types))

    def validate_type_dict(self, types):
        # 'return' isn't used for anything yet but it may be shown by Libdoc
        # in the future. Trying to be forward compatible.
        names = set(self._argspec.argument_names + ['return'])
        extra = [t for t in types if t not in names]
        if extra:
            raise DataError('Type information given to non-existing '
                            'argument%s %s.'
                            % (s(extra), seq2str(sorted(extra))))
        return types

    def convert_type_list_to_dict(self, types):
        names = self._argspec.argument_names
        if len(types) > len(names):
            raise DataError('Type information given to %d argument%s but '
                            'keyword has only %d argument%s.'
                            % (len(types), s(types), len(names), s(names)))
        return {name: value for name, value in zip(names, types) if value}



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import ctypes
import time
from threading import current_thread, Lock, Timer


class Timeout:

    def __init__(self, timeout, error):
        self._runner_thread_id = current_thread().ident
        self._timer = Timer(timeout, self._timed_out)
        self._error = error
        self._timeout_occurred = False
        self._finished = False
        self._lock = Lock()

    def execute(self, runnable):
        try:
            self._start_timer()
            try:
                result = runnable()
            finally:
                self._cancel_timer()
            self._wait_for_raised_timeout()
            return result
        finally:
            if self._timeout_occurred:
                raise self._error

    def _start_timer(self):
        self._timer.start()

    def _cancel_timer(self):
        with self._lock:
            self._finished = True
            self._timer.cancel()

    def _wait_for_raised_timeout(self):
        if self._timeout_occurred:
            while True:
                time.sleep(0)

    def _timed_out(self):
        with self._lock:
            if self._finished:
                return
            self._timeout_occurred = True
        self._raise_timeout()

    def _raise_timeout(self):
        # See, for example, http://tomerfiliba.com/recipes/Thread2/
        # for more information about using PyThreadState_SetAsyncExc
        tid = ctypes.c_long(self._runner_thread_id)
        error = ctypes.py_object(type(self._error))
        while ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, error) > 1:
            ctypes.pythonapi.PyThreadState_SetAsyncExc(tid, None)
            time.sleep(0)  # give time for other threads



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.model import Tags
from robot.result import Keyword as KeywordResult
from robot.variables import VariableAssignment

from .arguments import ArgumentSpec
from .statusreporter import StatusReporter


class UserErrorHandler:
    """Created if creating handlers fail. Running it raises DataError.

    The idea is not to raise DataError at processing time and prevent all
    tests in affected test case file from executing. Instead, UserErrorHandler
    is created and if it is ever run DataError is raised then.
    """
    supports_embedded_arguments = False

    def __init__(self, error, name, libname=None, source=None, lineno=None):
        """
        :param robot.errors.DataError error: Occurred error.
        :param str name: Name of the affected keyword.
        :param str libname: Name of the affected library or resource.
        :param str source: Path to the source file.
        :param int lineno: Line number of the failing keyword.
        """
        self.error = error
        self.name = name
        self.libname = libname
        self.source = source
        self.lineno = lineno
        self.arguments = ArgumentSpec()
        self.timeout = None
        self.tags = Tags()

    @property
    def longname(self):
        return f'{self.libname}.{self.name}' if self.libname else self.name

    @property
    def doc(self):
        return f'*Creating keyword failed:* {self.error}'

    @property
    def shortdoc(self):
        return self.doc.splitlines()[0]

    def create_runner(self, name, languages=None):
        return self

    def run(self, kw, context, run=True):
        result = KeywordResult(kwname=self.name,
                               libname=self.libname,
                               args=kw.args,
                               assign=tuple(VariableAssignment(kw.assign)),
                               type=kw.type)
        with StatusReporter(kw, result, context, run):
            if run:
                raise self.error

    dry_run = run



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.variables import contains_variable

from .typeconverters import TypeConverter


class ArgumentConverter:

    def __init__(self, argspec, converters, dry_run=False, languages=None):
        """:type argspec: :py:class:`robot.running.arguments.ArgumentSpec`"""
        self._argspec = argspec
        self._converters = converters
        self._dry_run = dry_run
        self._languages = languages

    def convert(self, positional, named):
        return self._convert_positional(positional), self._convert_named(named)

    def _convert_positional(self, positional):
        names = self._argspec.positional
        converted = [self._convert(name, value)
                     for name, value in zip(names, positional)]
        if self._argspec.var_positional:
            converted.extend(self._convert(self._argspec.var_positional, value)
                             for value in positional[len(names):])
        return converted

    def _convert_named(self, named):
        names = set(self._argspec.positional) | set(self._argspec.named_only)
        var_named = self._argspec.var_named
        return [(name, self._convert(name if name in names else var_named, value))
                for name, value in named]

    def _convert(self, name, value):
        spec = self._argspec
        if (spec.types is None
                or self._dry_run and contains_variable(value, identifiers='$@&%')):
            return value
        conversion_error = None
        # Don't convert None if argument has None as a default value.
        # Python < 3.11 adds None to type hints automatically when using None as
        # a default value which preserves None automatically. This code keeps
        # the same behavior also with newer Python versions.
        if value is None and name in spec.defaults and spec.defaults[name] is None:
            return value
        if name in spec.types:
            converter = TypeConverter.converter_for(spec.types[name], self._converters,
                                                    self._languages)
            if converter:
                try:
                    return converter.convert(name, value)
                except ValueError as err:
                    conversion_error = err
        if name in spec.defaults:
            converter = TypeConverter.converter_for(type(spec.defaults[name]),
                                                    languages=self._languages)
            if converter:
                try:
                    return converter.convert(name, value, explicit_type=False,
                                             strict=bool(conversion_error))
                except ValueError as err:
                    conversion_error = conversion_error or err
        if conversion_error:
            raise conversion_error
        return value



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from inspect import isclass, signature, Parameter
from typing import get_type_hints

from robot.errors import DataError
from robot.utils import is_string, split_from_equals
from robot.variables import is_assign, is_scalar_assign

from .argumentspec import ArgumentSpec


class ArgumentParser(ABC):

    def __init__(self, type='Keyword', error_reporter=None):
        self._type = type
        self._error_reporter = error_reporter

    @abstractmethod
    def parse(self, source, name=None):
        raise NotImplementedError

    def _report_error(self, error):
        if self._error_reporter:
            self._error_reporter(error)
        else:
            raise DataError(f'Invalid argument specification: {error}')


class PythonArgumentParser(ArgumentParser):

    def parse(self, handler, name=None):
        spec = ArgumentSpec(name, self._type)
        self._set_args(spec, handler)
        self._set_types(spec, handler)
        return spec

    def _set_args(self, spec, handler):
        try:
            sig = signature(handler)
        except ValueError:        # Can occur with C functions (incl. many builtins).
            spec.var_positional = 'args'
            return
        except TypeError as err:  # Occurs if handler isn't actually callable.
            raise DataError(str(err))
        parameters = list(sig.parameters.values())
        # `inspect.signature` drops `self` with bound methods and that's the case when
        # inspecting keywords. `__init__` is got directly from class (i.e. isn't bound)
        # so we need to handle that case ourselves.
        # Partial objects do not have __name__ at least in Python =< 3.10.
        if getattr(handler, '__name__', None) == '__init__':
            parameters = parameters[1:]
        setters = {
            Parameter.POSITIONAL_ONLY: spec.positional_only.append,
            Parameter.POSITIONAL_OR_KEYWORD: spec.positional_or_named.append,
            Parameter.VAR_POSITIONAL: lambda name: setattr(spec, 'var_positional', name),
            Parameter.KEYWORD_ONLY: spec.named_only.append,
            Parameter.VAR_KEYWORD: lambda name: setattr(spec, 'var_named', name),
        }
        for param in parameters:
            setters[param.kind](param.name)
            if param.default is not param.empty:
                spec.defaults[param.name] = param.default

    def _set_types(self, spec, handler):
        # If types are set using the `@keyword` decorator, use them. Including when
        # types are explicitly disabled with `@keyword(types=None)`. Otherwise read
        # type hints.
        if isclass(handler):
            handler = handler.__init__
        robot_types = getattr(handler, 'robot_types', ())
        if robot_types or robot_types is None:
            spec.types = robot_types
        else:
            spec.types = self._get_type_hints(handler)

    def _get_type_hints(self, handler):
        try:
            return get_type_hints(handler)
        except Exception:  # Can raise pretty much anything
            # Not all functions have `__annotations__`.
            # https://github.com/robotframework/robotframework/issues/4059
            return getattr(handler, '__annotations__', {})


class ArgumentSpecParser(ArgumentParser):

    def parse(self, argspec, name=None):
        spec = ArgumentSpec(name, self._type)
        named_only = positional_only_separator_seen = False
        for arg in argspec:
            arg = self._validate_arg(arg)
            if spec.var_named:
                self._report_error('Only last argument can be kwargs.')
            elif self._is_positional_only_separator(arg):
                if named_only:
                    self._report_error('Positional-only separator must be before '
                                       'named-only arguments.')
                if positional_only_separator_seen:
                    self._report_error('Too many positional-only separators.')
                spec.positional_only = spec.positional_or_named
                spec.positional_or_named = []
                positional_only_separator_seen = True
            elif isinstance(arg, tuple):
                arg, default = arg
                arg = self._add_arg(spec, arg, named_only)
                spec.defaults[arg] = default
            elif self._is_var_named(arg):
                spec.var_named = self._format_var_named(arg)
            elif self._is_var_positional(arg):
                if named_only:
                    self._report_error('Cannot have multiple varargs.')
                if not self._is_named_only_separator(arg):
                    spec.var_positional = self._format_var_positional(arg)
                named_only = True
            elif spec.defaults and not named_only:
                self._report_error('Non-default argument after default arguments.')
            else:
                self._add_arg(spec, arg, named_only)
        return spec

    @abstractmethod
    def _validate_arg(self, arg):
        raise NotImplementedError

    @abstractmethod
    def _is_var_named(self, arg):
        raise NotImplementedError

    @abstractmethod
    def _format_var_named(self, kwargs):
        raise NotImplementedError

    @abstractmethod
    def _is_positional_only_separator(self, arg):
        raise NotImplementedError

    @abstractmethod
    def _is_named_only_separator(self, arg):
        raise NotImplementedError

    @abstractmethod
    def _is_var_positional(self, arg):
        raise NotImplementedError

    @abstractmethod
    def _format_var_positional(self, varargs):
        raise NotImplementedError

    def _format_arg(self, arg):
        return arg

    def _add_arg(self, spec, arg, named_only=False):
        arg = self._format_arg(arg)
        target = spec.positional_or_named if not named_only else spec.named_only
        target.append(arg)
        return arg


class DynamicArgumentParser(ArgumentSpecParser):

    def _validate_arg(self, arg):
        if isinstance(arg, tuple):
            if self._is_invalid_tuple(arg):
                self._report_error(f'Invalid argument "{arg}".')
            if len(arg) == 1:
                return arg[0]
            return arg
        if '=' in arg:
            return tuple(arg.split('=', 1))
        return arg

    def _is_invalid_tuple(self, arg):
        return (len(arg) > 2
                or not is_string(arg[0])
                or (arg[0].startswith('*') and len(arg) > 1))

    def _is_var_named(self, arg):
        return arg[:2] == '**'

    def _format_var_named(self, kwargs):
        return kwargs[2:]

    def _is_var_positional(self, arg):
        return arg and arg[0] == '*'

    def _is_positional_only_separator(self, arg):
        return arg == '/'

    def _is_named_only_separator(self, arg):
        return arg == '*'

    def _format_var_positional(self, varargs):
        return varargs[1:]


class UserKeywordArgumentParser(ArgumentSpecParser):

    def _validate_arg(self, arg):
        arg, default = split_from_equals(arg)
        if not (is_assign(arg) or arg == '@{}'):
            self._report_error(f"Invalid argument syntax '{arg}'.")
        if default is None:
            return arg
        if not is_scalar_assign(arg):
            typ = 'list' if arg[0] == '@' else 'dictionary'
            self._report_error(f"Only normal arguments accept default values, "
                               f"{typ} arguments like '{arg}' do not.")
        return arg, default

    def _is_var_named(self, arg):
        return arg and arg[0] == '&'

    def _format_var_named(self, kwargs):
        return kwargs[2:-1]

    def _is_var_positional(self, arg):
        return arg and arg[0] == '@'

    def _is_positional_only_separator(self, arg):
        return False

    def _is_named_only_separator(self, arg):
        return arg == '@{}'

    def _format_var_positional(self, varargs):
        return varargs[2:-1]

    def _format_arg(self, arg):
        return arg[2:-1]



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Any, Callable, TypeVar


T = TypeVar('T', bound=Callable[..., Any])

# Type Alias for objects that are only known at runtime. This should be Used as a
# default value for generic classes that also use `@copy_signature` decorator
KnownAtRuntime = type(object)


def copy_signature(target: T) -> Callable[..., T]:
    """A decorator that applies the signature of `T` to any function that it decorates
    see https://github.com/python/typing/issues/270#issuecomment-555966301 for source
    and discussion.
    """
    def decorator(func):
        return func

    return decorator



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.errors import DataError
from robot.utils import get_timestamp, file_writer, seq2str2

from .logger import LOGGER
from .loggerhelper import IsLogged


def DebugFile(path):
    if not path:
        LOGGER.info('No debug file')
        return None
    try:
        outfile = file_writer(path, usage='debug')
    except DataError as err:
        LOGGER.error(err.message)
        return None
    else:
        LOGGER.info('Debug file: %s' % path)
        return _DebugFileWriter(outfile)


class _DebugFileWriter:
    _separators = {'SUITE': '=', 'TEST': '-', 'KEYWORD': '~'}

    def __init__(self, outfile):
        self._indent = 0
        self._kw_level = 0
        self._separator_written_last = False
        self._outfile = outfile
        self._is_logged = IsLogged('DEBUG')

    def start_suite(self, suite):
        self._separator('SUITE')
        self._start('SUITE', suite.longname)
        self._separator('SUITE')

    def end_suite(self, suite):
        self._separator('SUITE')
        self._end('SUITE', suite.longname, suite.elapsedtime)
        self._separator('SUITE')
        if self._indent == 0:
            LOGGER.output_file('Debug', self._outfile.name)
            self.close()

    def start_test(self, test):
        self._separator('TEST')
        self._start('TEST', test.name)
        self._separator('TEST')

    def end_test(self, test):
        self._separator('TEST')
        self._end('TEST', test.name, test.elapsedtime)
        self._separator('TEST')

    def start_keyword(self, kw):
        if self._kw_level == 0:
            self._separator('KEYWORD')
        self._start(kw.type, kw.name, kw.args)
        self._kw_level += 1

    def end_keyword(self, kw):
        self._end(kw.type, kw.name, kw.elapsedtime)
        self._kw_level -= 1

    def log_message(self, msg):
        if self._is_logged(msg.level):
            self._write(msg.message, level=msg.level, timestamp=msg.timestamp)

    def close(self):
        if not self._outfile.closed:
            self._outfile.close()

    def _start(self, type_, name, args=''):
        args = ' ' + seq2str2(args)
        self._write('+%s START %s: %s%s' % ('-'*self._indent, type_, name, args))
        self._indent += 1

    def _end(self, type_, name, elapsed):
        self._indent -= 1
        self._write('+%s END %s: %s (%s)' % ('-'*self._indent, type_, name, elapsed))

    def _separator(self, type_):
        self._write(self._separators[type_] * 78, separator=True)

    def _write(self, text, separator=False, level='INFO', timestamp=None):
        if separator and self._separator_written_last:
            return
        if not separator:
            text = '%s - %s - %s' % (timestamp or get_timestamp(), level, text)
        self._outfile.write(text.rstrip() + '\n')
        self._outfile.flush()
        self._separator_written_last = separator



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from robot.errors import DataError

from .dotted import DottedOutput
from .quiet import NoOutput, QuietOutput
from .verbose import VerboseOutput


def ConsoleOutput(type='verbose', width=78, colors='AUTO', markers='AUTO',
                  stdout=None, stderr=None):
    upper = type.upper()
    if upper == 'VERBOSE':
        return VerboseOutput(width, colors, markers, stdout, stderr)
    if upper == 'DOTTED':
        return DottedOutput(width, colors, stdout, stderr)
    if upper == 'QUIET':
        return QuietOutput(colors, stderr)
    if upper == 'NONE':
        return NoOutput()
    raise DataError("Invalid console output type '%s'. Available "
                    "'VERBOSE', 'DOTTED', 'QUIET' and 'NONE'." % type)



#  Copyright 2008-2015 Nokia Networks
#  Copyright 2016-     Robot Framework Foundation
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

"""Implements the core test execution logic.

The public API of this module consists of the following objects:

* :class:`~robot.running.model.TestSuite` for creating an executable
  test suite structure programmatically.

* :class:`~robot.running.builder.builders.TestSuiteBuilder` for creating
  executable test suites based on data on a file system.
  Instead of using this class directly, it is possible to use the
  :meth:`TestSuite.from_file_system <robot.running.model.TestSuite.from_file_system>`
  classmethod that uses it internally.

* Classes used by :class:`~robot.running.model.TestSuite`, such as
  :class:`~robot.running.model.TestCase`, :class:`~robot.running.model.Keyword`
  and :class:`~robot.running.model.If` that are defined in the
  :mod:`robot.running.model` module. These classes are typically only needed
  in type hints.

* :class:`~robot.running.builder.settings.TestDefaults` that is part of the
  `external parsing API`__ and also typically needed only in type hints.

__ http://robotframework.org/robotframework/latest/RobotFrameworkUserGuide.html#parser-interface

:class:`~robot.running.model.TestSuite` and
:class:`~robot.running.builder.builders.TestSuiteBuilder` can be imported also via
the :mod:`robot.api` package.

.. note:: Prior to Robot Framework 6.1, only some classes in
          :mod:`robot.running.model` were exposed via :mod:`robot.running`.

Examples
--------

First, let's assume we have the following test suite in file
``activate_skynet.robot``::

    *** Settings ***
    Library    OperatingSystem

    *** Test Cases ***
    Should Activate Skynet
        [Tags]    smoke
        [Setup]    Set Environment Variable    SKYNET    activated
        Environment Variable Should Be Set    SKYNET

We can easily create an executable test suite based on the above file::

    from robot.api import TestSuite

    suite = TestSuite.from_file_system('path/to/activate_skynet.robot')

That was easy. Let's next generate the same test suite from scratch::

    from robot.api import TestSuite

    suite = TestSuite('Activate Skynet')
    suite.resource.imports.library('OperatingSystem')
    test = suite.tests.create('Should Activate Skynet', tags=['smoke'])
    test.setup.config(name='Set Environment Variable', args=['SKYNET', 'activated'])
    test.body.create_keyword('Environment Variable Should Be Set', args=['SKYNET'])

Not that complicated either, especially considering the flexibility. Notice
that the suite created based on the file could also be edited further using
the same API.

Now that we have a test suite ready, let's :meth:`execute it
<robot.running.model.TestSuite.run>` and verify that the returned
:class:`~robot.result.executionresult.Result` object contains correct
information::

    result = suite.run(output='skynet.xml')

    assert result.return_code == 0
    assert result.suite.name == 'Activate Skynet'
    test = result.suite.tests[0]
    assert test.name == 'Should Activate Skynet'
    assert test.passed
    stats = result.suite.statistics
    assert stats.total == 1 and stats.passed == 1 and stats.failed == 0

Running the suite generates a normal output XML file, unless it is disabled
by using ``output=None``. Generating log, report, and xUnit files based on
the results is possible using the
:class:`~robot.reporting.resultwriter.ResultWriter` class::

    from robot.api import ResultWriter

    # Report and xUnit files can be generated based on the result object.
    ResultWriter(result).write_results(report='skynet.html', log=None)
    # Generating log files requires processing the earlier generated output XML.
    ResultWriter('skynet.xml').write_results()
"""

from .arguments import ArgInfo, ArgumentSpec, TypeConverter, TypeInfo
from .builder import ResourceFileBuilder, TestDefaults, TestSuiteBuilder
from .context import EXECUTION_CONTEXTS
from .model import (Break, Continue, Error, For, If, IfBranch, Keyword, Return,
                    ResourceFile, TestCase, TestSuite, Try, TryBranch, While)
from .runkwregister import RUN_KW_REGISTER
from .testlibraries import TestLibrary
from .usererrorhandler import UserErrorHandler
from .userkeyword import UserLibrary
